from typing import List, Tuple, Dict, Optional

import torch
import torch.nn as nn
import torch.nn.utils.rnn as rnn_utils

from pyhealth.datasets import SampleEHRDataset
from pyhealth.models import BaseModel
from pyhealth.models.utils import get_last_visit
import torch.nn.functional as F


class AgentLayer(nn.Module):
    """Dr. Agent layer.

    Paper: Junyi Gao et al. Dr. Agent: Clinical predictive model via mimicked second opinions. JAMIA.

    This layer is used in the Dr. Agent model. But it can also be used as a
    standalone layer.

    Args:
        input_dim: dynamic feature size.
        static_dim: static feature size, if 0, then no static feature is used.
        cell: rnn cell type. Default is "gru".
        use_baseline: whether to use baseline for the RL agent. Default is True.
        n_actions: number of historical visits to choose. Default is 10.
        n_units: number of hidden units in each agent. Default is 64.
        fusion_dim: number of hidden units in the final representation. Default is 128.
        n_hidden: number of hidden units in the rnn. Default is 128.
        dropout: dropout rate. Default is 0.5.
        lamda: weight for the agent selected hidden state and the current hidden state. Default is 0.5.

    Examples:
        >>> from pyhealth.models import AgentLayer
        >>> input = torch.randn(3, 128, 64)  # [batch size, sequence len, feature_size]
        >>> layer = AgentLayer(64)
        >>> c, _ = layer(input)
        >>> c.shape
        torch.Size([3, 128])
    """

    def __init__(
        self,
        input_dim: int,
        static_dim: int = 0,
        cell: str = "gru",
        use_baseline: bool = True,
        n_actions: int = 10,
        n_units: int = 64,
        n_hidden: int = 128,
        dropout: int = 0.5,
        lamda: int = 0.5,
    ):
        super(AgentLayer, self).__init__()

        if cell not in ["gru", "lstm"]:
            raise ValueError("Only gru and lstm are supported for cell.")

        self.cell = cell
        self.use_baseline = use_baseline
        self.n_actions = n_actions
        self.n_units = n_units
        self.input_dim = input_dim
        self.n_hidden = n_hidden
        # self.n_output = n_output
        self.dropout = dropout
        self.lamda = lamda
        self.fusion_dim = n_hidden
        self.static_dim = static_dim

        self.agent1_action = []
        self.agent1_prob = []
        self.agent1_entropy = []
        self.agent1_baseline = []
        self.agent2_action = []
        self.agent2_prob = []
        self.agent2_entropy = []
        self.agent2_baseline = []

        self.agent1_fc1 = nn.Linear(self.n_hidden + self.static_dim, self.n_units)
        self.agent2_fc1 = nn.Linear(self.input_dim + self.static_dim, self.n_units)
        self.agent1_fc2 = nn.Linear(self.n_units, self.n_actions)
        self.agent2_fc2 = nn.Linear(self.n_units, self.n_actions)
        if use_baseline == True:
            self.agent1_value = nn.Linear(self.n_units, 1)
            self.agent2_value = nn.Linear(self.n_units, 1)

        if self.cell == "lstm":
            self.rnn = nn.LSTMCell(self.input_dim, self.n_hidden)
        else:
            self.rnn = nn.GRUCell(self.input_dim, self.n_hidden)

        for name, param in self.rnn.named_parameters():
            if "bias" in name:
                nn.init.constant_(param, 0.0)
            elif "weight" in name:
                nn.init.orthogonal_(param)

        if dropout > 0.0:
            self.nn_dropout = nn.Dropout(p=dropout)
        if self.static_dim > 0:
            self.init_h = nn.Linear(self.static_dim, self.n_hidden)
            self.init_c = nn.Linear(self.static_dim, self.n_hidden)
            self.fusion = nn.Linear(self.n_hidden + self.static_dim, self.fusion_dim)
        # self.output = nn.Linear(self.fusion_dim, self.n_output)

        self.sigmoid = nn.Sigmoid()
        self.softmax = nn.Softmax(dim=1)
        self.tanh = nn.Tanh()
        self.relu = nn.ReLU()

    def choose_action(self, observation, agent=1):
        observation = observation.detach()

        if agent == 1:
            result_fc1 = self.agent1_fc1(observation)
            result_fc1 = self.tanh(result_fc1)
            result_fc2 = self.agent1_fc2(result_fc1)
            if self.use_baseline == True:
                result_value = self.agent1_value(result_fc1)
                self.agent1_baseline.append(result_value)
        else:
            result_fc1 = self.agent2_fc1(observation)
            result_fc1 = self.tanh(result_fc1)
            result_fc2 = self.agent2_fc2(result_fc1)
            if self.use_baseline == True:
                result_value = self.agent2_value(result_fc1)
                self.agent2_baseline.append(result_value)

        probs = self.softmax(result_fc2)
        m = torch.distributions.Categorical(probs)
        actions = m.sample()

        if agent == 1:
            self.agent1_entropy.append(m.entropy())
            self.agent1_action.append(actions.unsqueeze(-1))
            self.agent1_prob.append(m.log_prob(actions))
        else:
            self.agent2_entropy.append(m.entropy())
            self.agent2_action.append(actions.unsqueeze(-1))
            self.agent2_prob.append(m.log_prob(actions))

        return actions.unsqueeze(-1)

    def forward(
        self,
        x: torch.tensor,
        static: Optional[torch.tensor] = None,
        mask: Optional[torch.tensor] = None,
    ) -> Tuple[torch.tensor]:
        """Forward propagation.

        Args:
            x: a tensor of shape [batch size, sequence len, input_dim].
            static: a tensor of shape [batch size, static_dim].
            mask: an optional tensor of shape [batch size, sequence len], where
                1 indicates valid and 0 indicates invalid.

        Returns:
            last_output: a tensor of shape [batch size, n_hidden] representing the
                patient embedding.
            output: a tensor of shape [batch size, sequence len, n_hidden] representing the patient embedding at each time step.
        """
        # rnn will only apply dropout between layers

        batch_size = x.size(0)
        time_step = x.size(1)
        feature_dim = x.size(2)

        self.agent1_action = []
        self.agent1_prob = []
        self.agent1_entropy = []
        self.agent1_baseline = []
        self.agent2_action = []
        self.agent2_prob = []
        self.agent2_entropy = []
        self.agent2_baseline = []

        if self.static_dim > 0:
            cur_h = self.init_h(static)
            if self.cell == "lstm":
                cur_c = self.init_c(static)
        else:
            cur_h = torch.zeros(
                batch_size, self.n_hidden, dtype=torch.float32, device=x.device
            )
            if self.cell == "lstm":
                cur_c = torch.zeros(
                    batch_size, self.n_hidden, dtype=torch.float32, device=x.device
                )

        h = []
        for cur_time in range(time_step):
            cur_input = x[:, cur_time, :]

            if cur_time == 0:
                obs_1 = cur_h
                obs_2 = cur_input

                if self.static_dim > 0:
                    obs_1 = torch.cat((obs_1, static), dim=1)
                    obs_2 = torch.cat((obs_2, static), dim=1)

                self.choose_action(obs_1, 1).long()
                self.choose_action(obs_2, 2).long()

                observed_h = (
                    torch.zeros_like(cur_h, dtype=torch.float32)
                    .view(-1)
                    .repeat(self.n_actions)
                    .view(self.n_actions, batch_size, self.n_hidden)
                )
                action_h = cur_h
                if self.cell == "lstm":
                    observed_c = (
                        torch.zeros_like(cur_c, dtype=torch.float32)
                        .view(-1)
                        .repeat(self.n_actions)
                        .view(self.n_actions, batch_size, self.n_hidden)
                    )
                    action_c = cur_c

            else:
                observed_h = torch.cat((observed_h[1:], cur_h.unsqueeze(0)), 0)

                obs_1 = observed_h.mean(dim=0)
                obs_2 = cur_input

                if self.static_dim > 0:
                    obs_1 = torch.cat((obs_1, static), dim=1)
                    obs_2 = torch.cat((obs_2, static), dim=1)

                act_idx1 = self.choose_action(obs_1, 1).long()
                act_idx2 = self.choose_action(obs_2, 2).long()
                batch_idx = torch.arange(batch_size, dtype=torch.long).unsqueeze(-1)
                action_h1 = observed_h[act_idx1, batch_idx, :].squeeze(1)
                action_h2 = observed_h[act_idx2, batch_idx, :].squeeze(1)
                action_h = (action_h1 + action_h2) / 2
                if self.cell == "lstm":
                    observed_c = torch.cat((observed_c[1:], cur_c.unsqueeze(0)), 0)
                    action_c1 = observed_c[act_idx1, batch_idx, :].squeeze(1)
                    action_c2 = observed_c[act_idx2, batch_idx, :].squeeze(1)
                    action_c = (action_c1 + action_c2) / 2

            if self.cell == "lstm":
                weighted_h = self.lamda * action_h + (1 - self.lamda) * cur_h
                weighted_c = self.lamda * action_c + (1 - self.lamda) * cur_c
                rnn_state = (weighted_h, weighted_c)
                cur_h, cur_c = self.rnn(cur_input, rnn_state)
            else:
                weighted_h = self.lamda * action_h + (1 - self.lamda) * cur_h
                cur_h = self.rnn(cur_input, weighted_h)
            h.append(cur_h)

        h = torch.stack(h, dim=1)

        if self.static_dim > 0:
            static = static.unsqueeze(1).repeat(1, time_step, 1)
            h = torch.cat((h, static), dim=2)
            h = self.fusion(h)

        last_out = get_last_visit(h, mask)

        if self.dropout > 0.0:
            last_out = self.nn_dropout(last_out)
        return last_out, h


class Agent(BaseModel):
    """Dr. Agent model.

    Paper: Junyi Gao et al. Dr. Agent: Clinical predictive model via mimicked second opinions. JAMIA.

    Note:
        We use separate Dr. Agent layers for different feature_keys.
        Currently, we automatically support different input formats:
            - code based input (need to use the embedding table later)
            - float/int based value input
        We follow the current convention for the Dr. Agent model:
            - case 1. [code1, code2, code3, ...]
                - we will assume the code follows the order; our model will encode
                each code into a vector and apply Dr. Agent on the code level
            - case 2. [[code1, code2]] or [[code1, code2], [code3, code4, code5], ...]
                - we will assume the inner bracket follows the order; our model first
                use the embedding table to encode each code into a vector and then use
                average/mean pooling to get one vector for one inner bracket; then use
                Dr. Agent one the braket level
            - case 3. [[1.5, 2.0, 0.0]] or [[1.5, 2.0, 0.0], [8, 1.2, 4.5], ...]
                - this case only makes sense when each inner bracket has the same length;
                we assume each dimension has the same meaning; we run Dr. Agent directly
                on the inner bracket level, similar to case 1 after embedding table
            - case 4. [[[1.5, 2.0, 0.0]]] or [[[1.5, 2.0, 0.0], [8, 1.2, 4.5]], ...]
                - this case only makes sense when each inner bracket has the same length;
                we assume each dimension has the same meaning; we run Dr. Agent directly
                on the inner bracket level, similar to case 2 after embedding table

    Args:
        dataset: the dataset to train the model. It is used to query certain
            information such as the set of all tokens.
        feature_keys:  list of keys in samples to use as features,
            e.g. ["conditions", "procedures"].
        label_key: key in samples to use as label (e.g., "drugs").
        mode: one of "binary", "multiclass", or "multilabel".
        static_keys: the key in samples to use as static features, e.g. "demographics". Default is None.
                     we only support numerical static features.
        embedding_dim: the embedding dimension. Default is 128.
        hidden_dim: the hidden dimension of the RNN in the Dr. Agent layer. Default is 128.
        use_baseline: whether to use the baseline value to calculate the RL loss. Default is True.
        **kwargs: other parameters for the Dr. Agent layer.


    Examples:
        >>> from pyhealth.datasets import SampleEHRDataset
        >>> samples = [
        ...         {
        ...             "patient_id": "patient-0",
        ...             "visit_id": "visit-0",
        ...             "list_codes": ["505800458", "50580045810", "50580045811"],  # NDC
        ...             "list_vectors": [[1.0, 2.55, 3.4], [4.1, 5.5, 6.0]],
        ...             "list_list_codes": [["A05B", "A05C", "A06A"], ["A11D", "A11E"]],  # ATC-4
        ...             "list_list_vectors": [
        ...                 [[1.8, 2.25, 3.41], [4.50, 5.9, 6.0]],
        ...                 [[7.7, 8.5, 9.4]],
        ...             ],
        ...             "demographic": [0.0, 2.0, 1.5],
        ...             "label": 1,
        ...         },
        ...         {
        ...             "patient_id": "patient-0",
        ...             "visit_id": "visit-1",
        ...             "list_codes": [
        ...                 "55154191800",
        ...                 "551541928",
        ...                 "55154192800",
        ...                 "705182798",
        ...                 "70518279800",
        ...             ],
        ...             "list_vectors": [[1.4, 3.2, 3.5], [4.1, 5.9, 1.7], [4.5, 5.9, 1.7]],
        ...             "list_list_codes": [["A04A", "B035", "C129"]],
        ...             "list_list_vectors": [
        ...                 [[1.0, 2.8, 3.3], [4.9, 5.0, 6.6], [7.7, 8.4, 1.3], [7.7, 8.4, 1.3]],
        ...             ],
        ...             "demographic": [0.0, 2.0, 1.5],
        ...             "label": 0,
        ...         },
        ...     ]
        >>> dataset = SampleEHRDataset(samples=samples, dataset_name="test")
        >>>
        >>> from pyhealth.models import Agent
        >>> model = Agent(
        ...         dataset=dataset,
        ...         feature_keys=[
        ...             "list_codes",
        ...             "list_vectors",
        ...             "list_list_codes",
        ...             "list_list_vectors",
        ...         ],
        ...         label_key="label",
        ...         static_key="demographic",
        ...         mode="binary"
        ...     )
        >>>
        >>> from pyhealth.datasets import get_dataloader
        >>> train_loader = get_dataloader(dataset, batch_size=2, shuffle=True)
        >>> data_batch = next(iter(train_loader))
        >>>
        >>> ret = model(**data_batch)
        >>> print(ret)
        {
            'loss': tensor(1.4059, grad_fn=<AddBackward0>),
            'y_prob': tensor([[0.4861], [0.5348]], grad_fn=<SigmoidBackward0>),
            'y_true': tensor([[0.], [1.]]),
            'logit': tensor([[-0.0556], [0.1392]], grad_fn=<AddmmBackward0>)
        }
        >>>

    """

    def __init__(
        self,
        dataset: SampleEHRDataset,
        feature_keys: List[str],
        label_key: str,
        mode: str,
        static_key: Optional[str] = None,
        embedding_dim: int = 128,
        hidden_dim: int = 128,
        use_baseline: bool = True,
        **kwargs,
    ):
        super(Agent, self).__init__(
            dataset=dataset,
            feature_keys=feature_keys,
            label_key=label_key,
            mode=mode,
        )
        self.embedding_dim = embedding_dim
        self.hidden_dim = hidden_dim

        # validate kwargs for Dr. Agent layer
        if "feature_size" in kwargs:
            raise ValueError("feature_size is determined by embedding_dim")

        # the key of self.feat_tokenizers only contains the code based inputs
        self.feat_tokenizers = {}
        self.static_key = static_key
        self.use_baseline = use_baseline
        self.label_tokenizer = self.get_label_tokenizer()
        # the key of self.embeddings only contains the code based inputs
        self.embeddings = nn.ModuleDict()
        # the key of self.linear_layers only contains the float/int based inputs
        self.linear_layers = nn.ModuleDict()

        self.static_dim = 0
        if self.static_key is not None:
            self.static_dim = self.dataset.input_info[self.static_key]["len"]

        self.agent = nn.ModuleDict()
        # add feature Dr. Agent layers
        for feature_key in self.feature_keys:
            input_info = self.dataset.input_info[feature_key]
            # sanity check
            if input_info["type"] not in [str, float, int]:
                raise ValueError(
                    "Dr. Agent only supports str code, float and int as input types"
                )
            elif (input_info["type"] == str) and (input_info["dim"] not in [2, 3]):
                raise ValueError(
                    "Dr. Agent only supports 2-dim or 3-dim str code as input types"
                )
            elif (input_info["type"] in [float, int]) and (
                input_info["dim"] not in [2, 3]
            ):
                raise ValueError(
                    "Dr. Agent only supports 2-dim or 3-dim float and int as input types"
                )

            # for code based input, we need Type
            # for float/int based input, we need Type, input_dim
            self.add_feature_transform_layer(feature_key, input_info)
            self.agent[feature_key] = AgentLayer(
                input_dim=embedding_dim,
                static_dim=self.static_dim,
                n_hidden=hidden_dim,
                **kwargs,
            )

        output_size = self.get_output_size(self.label_tokenizer)
        self.fc = nn.Linear(len(self.feature_keys) * self.hidden_dim, output_size)

    def get_loss(self, model, pred, true, mask, gamma=0.9, entropy_term=0.01):

        if self.mode == "binary":
            pred = torch.sigmoid(pred)
            rewards = ((pred - 0.5) * 2 * true).squeeze()
        elif self.mode == "multiclass":
            pred = torch.softmax(pred, dim=-1)
            y_onehot = torch.zeros_like(pred).scatter(1, true.unsqueeze(1), 1)
            rewards = (pred * y_onehot).sum(-1).squeeze()
        elif self.mode == "multilabel":
            pred = torch.sigmoid(pred)
            rewards = (
                ((pred - 0.5) * 2 * true).sum(dim=-1) / (true.sum(dim=-1) + 1e-7)
            ).squeeze()
        elif self.mode == "regression":
            rewards = (1 / torch.abs(pred - true)).squeeze()  # b*t
            rewards = torch.clamp(rewards, min=0, max=5)
        else:
            raise ValueError(
                "mode should be binary, multiclass, multilabel or regression"
            )

        act_prob1 = model.agent1_prob
        act_prob1 = torch.stack(act_prob1).permute(1, 0).to(self.device)
        act_prob1 = act_prob1 * mask.view(act_prob1.size(0), act_prob1.size(1))
        act_entropy1 = model.agent1_entropy
        act_entropy1 = torch.stack(act_entropy1).permute(1, 0).to(self.device)
        act_entropy1 = act_entropy1 * mask.view(
            act_entropy1.size(0), act_entropy1.size(1)
        )
        if self.use_baseline == True:
            act_baseline1 = model.agent1_baseline
            act_baseline1 = (
                torch.stack(act_baseline1).squeeze(-1).permute(1, 0).to(self.device)
            )
            act_baseline1 = act_baseline1 * mask.view(
                act_baseline1.size(0), act_baseline1.size(1)
            )

        act_prob2 = model.agent2_prob
        act_prob2 = torch.stack(act_prob2).permute(1, 0).to(self.device)
        act_prob2 = act_prob2 * mask.view(act_prob2.size(0), act_prob2.size(1))
        act_entropy2 = model.agent2_entropy
        act_entropy2 = torch.stack(act_entropy2).permute(1, 0).to(self.device)
        act_entropy2 = act_entropy2 * mask.view(
            act_entropy2.size(0), act_entropy2.size(1)
        )
        if self.use_baseline == True:
            act_baseline2 = model.agent2_baseline
            act_baseline2 = (
                torch.stack(act_baseline2).squeeze(-1).permute(1, 0).to(self.device)
            )
            act_baseline2 = act_baseline2 * mask.view(
                act_baseline2.size(0), act_baseline2.size(1)
            )

        running_rewards = []
        discounted_rewards = 0
        for i in reversed(range(act_prob1.size(1))):
            if i == act_prob1.size(1) - 1:
                discounted_rewards = rewards + gamma * discounted_rewards
            else:
                discounted_rewards = (
                    torch.zeros_like(rewards) + gamma * discounted_rewards
                )
            running_rewards.insert(0, discounted_rewards)
        rewards = torch.stack(running_rewards).permute(1, 0)
        # rewards = (rewards - rewards.mean(dim=1).unsqueeze(-1)) / (
        #     rewards.std(dim=1) + 1e-7
        # ).unsqueeze(-1)
        rewards = rewards.detach()

        if self.use_baseline == True:
            loss_value1 = torch.sum((rewards - act_baseline1) ** 2, dim=1) / torch.sum(
                mask, dim=1
            )
            loss_value1 = torch.mean(loss_value1)
            loss_value2 = torch.sum((rewards - act_baseline2) ** 2, dim=1) / torch.sum(
                mask, dim=1
            )
            loss_value2 = torch.mean(loss_value2)
            loss_value = loss_value1 + loss_value2
            loss_RL1 = -torch.sum(
                act_prob1 * (rewards - act_baseline1) + entropy_term * act_entropy1,
                dim=1,
            ) / torch.sum(mask, dim=1)
            loss_RL1 = torch.mean(loss_RL1)
            loss_RL2 = -torch.sum(
                act_prob2 * (rewards - act_baseline2) + entropy_term * act_entropy2,
                dim=1,
            ) / torch.sum(mask, dim=1)
            loss_RL2 = torch.mean(loss_RL2)
            loss_RL = loss_RL1 + loss_RL2
            loss = loss_RL + loss_value
        else:
            loss_RL1 = -torch.sum(
                act_prob1 * rewards + entropy_term * act_entropy1, dim=1
            ) / torch.sum(mask, dim=1)
            loss_RL1 = torch.mean(loss_RL1)
            loss_RL2 = -torch.sum(
                act_prob2 * rewards + entropy_term * act_entropy2, dim=1
            ) / torch.sum(mask, dim=1)
            loss_RL2 = torch.mean(loss_RL2)
            loss_RL = loss_RL1 + loss_RL2
            loss = loss_RL

        return loss

    def forward(self, **kwargs) -> Dict[str, torch.Tensor]:
        """Forward propagation.

        The label `kwargs[self.label_key]` is a list of labels for each patient.

        Args:
            **kwargs: keyword arguments for the model. The keys must contain
                all the feature keys and the label key.

        Returns:
            A dictionary with the following keys:
                loss: a scalar tensor representing the final loss.
                loss_task: a scalar tensor representing the task loss.
                loss_RL: a scalar tensor representing the RL loss.
                y_prob: a tensor representing the predicted probabilities.
                y_true: a tensor representing the true labels.
        """
        patient_emb = []
        mask_dict = {}
        for feature_key in self.feature_keys:
            input_info = self.dataset.input_info[feature_key]
            dim_, type_ = input_info["dim"], input_info["type"]

            # for case 1: [code1, code2, code3, ...]
            if (dim_ == 2) and (type_ == str):
                x = self.feat_tokenizers[feature_key].batch_encode_2d(
                    kwargs[feature_key]
                )
                # (patient, event)
                x = torch.tensor(x, dtype=torch.long, device=self.device)
                # (patient, event, embedding_dim)
                x = self.embeddings[feature_key](x)
                # (patient, event)
                mask = torch.any(x !=0, dim=2)
                mask_dict[feature_key] = mask

            # for case 2: [[code1, code2], [code3, ...], ...]
            elif (dim_ == 3) and (type_ == str):
                x = self.feat_tokenizers[feature_key].batch_encode_3d(
                    kwargs[feature_key]
                )
                # (patient, visit, event)
                x = torch.tensor(x, dtype=torch.long, device=self.device)
                # (patient, visit, event, embedding_dim)
                x = self.embeddings[feature_key](x)
                # (patient, visit, embedding_dim)
                x = torch.sum(x, dim=2)
                # (patient, visit)
                mask = torch.any(x !=0, dim=2)
                mask_dict[feature_key] = mask

            # for case 3: [[1.5, 2.0, 0.0], ...]
            elif (dim_ == 2) and (type_ in [float, int]):
                x, mask = self.padding2d(kwargs[feature_key])
                # (patient, event, values)
                x = torch.tensor(x, dtype=torch.float, device=self.device)
                # (patient, event, embedding_dim)
                x = self.linear_layers[feature_key](x)
                # (patient, event)
                mask = mask.bool().to(self.device)
                mask_dict[feature_key] = mask

            # for case 4: [[[1.5, 2.0, 0.0], [1.8, 2.4, 6.0]], ...]
            elif (dim_ == 3) and (type_ in [float, int]):
                x, mask = self.padding3d(kwargs[feature_key])
                # (patient, visit, event, values)
                x = torch.tensor(x, dtype=torch.float, device=self.device)
                # (patient, visit, embedding_dim)
                x = torch.sum(x, dim=2)
                x = self.linear_layers[feature_key](x)
                # (patient, event)
                mask = mask[:, :, 0]
                mask = mask.bool().to(self.device)
                mask_dict[feature_key] = mask

            else:
                raise NotImplementedError

            if self.static_dim > 0:
                static = torch.tensor(
                    kwargs[self.static_key], dtype=torch.float, device=self.device
                )
                x, _ = self.agent[feature_key](x, static=static, mask=mask)
            else:
                x, _ = self.agent[feature_key](x, mask=mask)
            patient_emb.append(x)

        patient_emb = torch.cat(patient_emb, dim=1)
        # (patient, label_size)
        logits = self.fc(patient_emb)
        # obtain y_true, loss, y_prob
        y_true = self.prepare_labels(kwargs[self.label_key], self.label_tokenizer)
        loss_task = self.get_loss_function()(logits, y_true)

        loss_rl = 0
        for feature_key in self.feature_keys:
            cur_loss = self.get_loss(
                self.agent[feature_key], logits, y_true, mask_dict[feature_key]
            )
            loss_rl += cur_loss
        loss = loss_task + loss_rl
        y_prob = self.prepare_y_prob(logits)
        results = {
            "loss": loss,
            "y_prob": y_prob,
            "y_true": y_true,
            "logit": logits,
        }
        if kwargs.get("embed", False):
            results["embed"] = patient_emb
        return results


if __name__ == "__main__":
    from pyhealth.datasets import SampleEHRDataset

    samples = [
        {
            "patient_id": "patient-0",
            "visit_id": "visit-0",
            # "single_vector": [1, 2, 3],
            "list_codes": ["505800458", "50580045810", "50580045811"],  # NDC
            "list_vectors": [[1.0, 2.55, 3.4], [4.1, 5.5, 6.0]],
            "list_list_codes": [["A05B", "A05C", "A06A"], ["A11D", "A11E"]],  # ATC-4
            "list_list_vectors": [
                [[1.8, 2.25, 3.41], [4.50, 5.9, 6.0]],
                [[7.7, 8.5, 9.4]],
            ],
            "label": 1,
            "demographic": [1.0, 2.0, 1.3],
        },
        {
            "patient_id": "patient-0",
            "visit_id": "visit-1",
            # "single_vector": [1, 5, 8],
            "list_codes": [
                "55154191800",
                "551541928",
                "55154192800",
                "705182798",
                "70518279800",
            ],
            "list_vectors": [[1.4, 3.2, 3.5], [4.1, 5.9, 1.7], [4.5, 5.9, 1.7]],
            "list_list_codes": [["A04A", "B035", "C129"]],
            "list_list_vectors": [
                [[1.0, 2.8, 3.3], [4.9, 5.0, 6.6], [7.7, 8.4, 1.3], [7.7, 8.4, 1.3]],
            ],
            "label": 0,
            "demographic": [1.0, 2.0, 1.3],
        },
    ]

    # dataset
    dataset = SampleEHRDataset(samples=samples, dataset_name="test")

    # data loader
    from pyhealth.datasets import get_dataloader

    train_loader = get_dataloader(dataset, batch_size=2, shuffle=True)

    # model
    model = Agent(
        dataset=dataset,
        feature_keys=[
            "list_codes",
            "list_vectors",
            "list_list_codes",
            # "list_list_vectors",
        ],
        static_key="demographic",
        label_key="label",
        mode="binary",
    )

    # data batch
    data_batch = next(iter(train_loader))

    # try the model
    ret = model(**data_batch)
    print(ret)

    # try loss backward
    ret["loss"].backward()



import random
from typing import Dict

import numpy as np
import tqdm
from torch.utils.data import DataLoader


def convert_to_ir_format(samples):
    corpus = {}
    queries = {}
    qrels = {}
    corpus_meta = {}
    queries_meta = {}
    for sample in samples:
        corpus[sample["d_visit_id"]] = sample["d_conditions"]
        queries[sample["visit_id"]] = sample["conditions"]
        qrels[sample["visit_id"]] = {sample["d_visit_id"]: 1}
        corpus_meta[sample["d_visit_id"]] = {"age": sample["d_age"],
                                             "identifiers": sample["d_identifiers"]}
        queries_meta[sample["visit_id"]] = {"age": sample["age"],
                                            "identifiers": sample["identifiers"]}
    return corpus, queries, qrels, corpus_meta, queries_meta


def generate_candidates(corpus_meta, queries_meta):
    candidates = {}
    for q_id, q_meta in queries_meta.items():
        age = q_meta["age"]
        identifiers = q_meta["identifiers"]
        matches = []
        for d_id, d_meta in corpus_meta.items():
            if (d_meta["age"] <= age) and (d_meta["identifiers"] == identifiers):
                matches.append(d_id)
        candidates[q_id] = matches
    # solve the problem of empty candidates
    average_matches = int(np.mean([len(v) for v in candidates.values()]))
    for q_id, q_meta in queries_meta.items():
        if len(candidates[q_id]) == 0:
            # random select average_matches candidates
            candidates[q_id] = random.sample(list(corpus_meta.keys()), average_matches)
    return candidates


def filter_by_candidates(results, qrels, candidates):
    filtered_results = {}
    for q_id, scores in results.items():
        c_ids = list(qrels[q_id].keys())
        candidate_ids = candidates[q_id]
        for c_id in c_ids:
            if c_id not in candidate_ids:
                candidate_ids.append(c_id)
        filtered_results[q_id] = {c_id: scores[c_id] for c_id in candidate_ids}
    return filtered_results


def tvt_split(queries, qrels, train_ratio=0.7, val_ratio=0.1, test_ratio=0.2):
    assert train_ratio + val_ratio + test_ratio == 1
    qids = list(queries.keys())
    np.random.shuffle(qids)
    s1 = int(len(qids) * train_ratio)
    s2 = int(len(qids) * (train_ratio + val_ratio))
    train_qids = qids[:s1]
    val_qids = qids[s1:s2]
    test_qids = qids[s2:]
    train_queries = {q_id: queries[q_id] for q_id in train_qids}
    val_queries = {q_id: queries[q_id] for q_id in val_qids}
    test_queries = {q_id: queries[q_id] for q_id in test_qids}
    train_qrels = {q_id: qrels[q_id] for q_id in train_qids}
    val_qrels = {q_id: qrels[q_id] for q_id in val_qids}
    test_qrels = {q_id: qrels[q_id] for q_id in test_qids}
    return train_queries, val_queries, test_queries, train_qrels, val_qrels, test_qrels


def get_bm25_hard_negatives(bm25_model, corpus, queries, qrels):
    qrels_w_neg = {}
    for q_id, q in tqdm.tqdm(queries.items()):
        d_ids = [d_id for d_id in qrels[q_id] if qrels[q_id][d_id] > 0]
        ds = [corpus[d_id] for d_id in d_ids]
        for d_id, d in zip(d_ids, ds):
            scores = bm25_model.get_scores(d)
            for (ned_d_id, neg_s) in sorted(scores.items(), key=lambda x: x[1],
                                            reverse=True):
                if ned_d_id != d_id:
                    qrels_w_neg[q_id] = {d_id: 1, ned_d_id: -1}
                    break
    return qrels_w_neg


def collate_fn(samples):
    outputs = {k: [] for k in samples[0].keys()}
    for sample in samples:
        for k, v in sample.items():
            outputs[k].append(v)
    return outputs


def get_train_dataloader(
    corpus: Dict[str, Dict[str, str]],
    queries: Dict[str, str],
    qrels: Dict[str, Dict[str, int]],
    batch_size: int,
    shuffle: bool = True
):
    query_ids = list(queries.keys())
    train_samples = []
    for query_id in query_ids:
        s_q = queries[query_id]
        id_p, s_p, s_n = None, None, None
        assert len(qrels[query_id]) <= 2
        for corpus_id, score in qrels[query_id].items():
            if score == 1:
                id_p = corpus_id
                s_p = corpus[corpus_id]
            if score == -1:
                s_n = corpus[corpus_id]
        if s_n is not None:
            train_samples.append(
                {
                    "query_id": query_id,
                    "id_p": id_p,
                    "s_q": s_q,
                    "s_p": s_p,
                    "s_n": s_n,
                }
            )
        else:
            train_samples.append(
                {
                    "query_id": query_id,
                    "id_p": id_p,
                    "s_q": s_q,
                    "s_p": s_p,
                }
            )
    print("Loaded {} training pairs.".format(len(train_samples)))
    train_dataloader = DataLoader(
        train_samples, shuffle=shuffle, batch_size=batch_size, collate_fn=collate_fn
    )
    return train_dataloader


def get_eval_dataloader(
    corpus: Dict[str, Dict[str, str]],
    queries: Dict[str, str],
    batch_size: int
):
    corpus_ids = list(corpus.keys())
    eval_samples = []
    for corpus_id in corpus_ids:
        s = corpus[corpus_id]
        eval_samples.append(
            {
                "corpus_id": corpus_id,
                "s": s,
            }
        )
    print("Loaded {} eval corpus.".format(len(eval_samples)))
    eval_corpus_dataloader = DataLoader(
        eval_samples, shuffle=False, batch_size=batch_size, collate_fn=collate_fn
    )

    query_ids = list(queries.keys())
    eval_samples = []
    for query_id in query_ids:
        s = queries[query_id]
        eval_samples.append(
            {
                "query_id": query_id,
                "s": s,
            }
        )
    print("Loaded {} eval queries.".format(len(eval_samples)))
    eval_queries_dataloader = DataLoader(
        eval_samples, batch_size=batch_size, collate_fn=collate_fn
    )
    return eval_corpus_dataloader, eval_queries_dataloader



import os

import numpy as np

from pyhealth.datasets import BaseSignalDataset


class TUABDataset(BaseSignalDataset):
    """Base EEG dataset for the TUH Abnormal EEG Corpus

    Dataset is available at https://isip.piconepress.com/projects/tuh_eeg/html/downloads.shtml

    The TUAB dataset (or Temple University Hospital EEG Abnormal Corpus) is a collection of EEG data acquired at the Temple University Hospital. 
    
    The dataset contains both normal and abnormal EEG readings.

    Files are named in the form aaaaamye_s001_t000.edf. This includes the subject identifier ("aaaaamye"), the session number ("s001") and a token number ("t000"). EEGs are split into a series of files starting with *t000.edf, *t001.edf, ...

    Args:
        dataset_name: name of the dataset.
        root: root directory of the raw data. *You can choose to use the path to Cassette portion or the Telemetry portion.*
        dev: whether to enable dev mode (only use a small subset of the data).
            Default is False.
        refresh_cache: whether to refresh the cache; if true, the dataset will
            be processed from scratch and the cache will be updated. Default is False.

    Attributes:
        task: Optional[str], name of the task (e.g., "EEG_abnormal").
            Default is None.
        samples: Optional[List[Dict]], a list of samples, each sample is a dict with
            patient_id, record_id, and other task-specific attributes as key.
            Default is None.
        patient_to_index: Optional[Dict[str, List[int]]], a dict mapping patient_id to
            a list of sample indices. Default is None.
        visit_to_index: Optional[Dict[str, List[int]]], a dict mapping visit_id to a
            list of sample indices. Default is None.

    Examples:
        >>> from pyhealth.datasets import TUABDataset
        >>> dataset = TUABDataset(
        ...         root="/srv/local/data/TUH/tuh_eeg_abnormal/v3.0.0/edf/",
        ...     )
        >>> dataset.stat()
        >>> dataset.info()
    """

    def process_EEG_data(self):
        # create a map for data sets for latter mapping patients
        data_map = {
            "train/abnormal": "0",
            "train/normal": "1",
            "eval/abnormal": "2",
            "eval/normal": "3",
        }

        data_map_reverse = {
            "0": "train/abnormal",
            "1": "train/normal",
            "2": "eval/abnormal",
            "3": "eval/normal",
        }

        # get all file names
        all_files = {}

        train_abnormal_files = os.listdir(os.path.join(self.root, "train/abnormal/01_tcp_ar"))
        all_files["train/abnormal"] = train_abnormal_files

        train_normal_files = os.listdir(os.path.join(self.root, "train/normal/01_tcp_ar"))
        all_files["train/normal"] = train_normal_files

        eval_abnormal_files = os.listdir(os.path.join(self.root, "eval/abnormal/01_tcp_ar"))
        all_files["eval/abnormal"] = eval_abnormal_files

        eval_normal_files = os.listdir(os.path.join(self.root, "eval/normal/01_tcp_ar"))
        all_files["eval/normal"] = eval_normal_files


        # get all patient ids
        patient_ids = []
        for field, sub_data in all_files.items():
            patient_ids.extend(["{}_{}".format(data_map[field], data.split("_")[0]) for data in sub_data])

        patient_ids = list(set(patient_ids))

        if self.dev:
            patient_ids = patient_ids[:5]

        # get patient to record maps
        #    - key: pid:
        #    - value: [{"load_from_path": None, "patient_id": None, "signal_file": None, "label_file": None, "save_to_path": None}, ...]
        patients = {
            pid: []
            for pid in patient_ids
        }
           
        for pid in patient_ids:
            data_field = data_map_reverse[pid.split("_")[0]]
            patient_visits = [file for file in all_files[data_field] if file.split("_")[0] == pid.split("_")[1]]
            
            for visit in patient_visits:
                patients[pid].append({
                    "load_from_path": os.path.join(self.root, data_field, "01_tcp_ar"),
                    "patient_id": pid,
                    "visit_id": visit.strip(".edf").strip(pid.split("_")[1])[1:],
                    "signal_file": visit,
                    "label_file": visit,
                    "save_to_path": self.filepath,
                })
        
        return patients


if __name__ == "__main__":
    dataset = TUABDataset(
        root="/srv/local/data/TUH/tuh_eeg_abnormal/v3.0.0/edf/",
        dev=True,
        refresh_cache=True,
    )
    dataset.stat()
    dataset.info()
    print(list(dataset.patients.items())[0])



import logging
import os
from abc import ABC

from tqdm import tqdm
import pandas as pd
from pandarallel import pandarallel
from typing import Callable, Optional
from pyhealth.datasets.utils import MODULE_CACHE_PATH, hash_str
from pyhealth.medcode.pretrained_embeddings.kg_emb.datasets import SampleKGDataset
from pyhealth.utils import load_pickle, save_pickle


logger = logging.getLogger(__name__)

INFO_MSG = """
dataset.triples:
    Array((<head_entity>, <relation>, <tail_entity>))
"""


class BaseKGDataset(ABC):
    """Abstract base Knowledge Graph class
    
    This abstract class defines a uniform

    Each specific dataset will be a subclass of this abstract class, which can then
    be converted to samples dataset for different tasks by calling `self.set_task()`.

    Args:
        dataset_name: name of the dataset.
        root: root directory of the raw data (should contain many csv files).
        dev: whether to enable dev mode (only use a small subset of the data).
            Default is False.
        refresh_cache: whether to refresh the cache; if true, the dataset will
            be processed from scratch and the cache will be updated. Default is False.
    
    """

    def __init__(
        self,
        root: str,
        dataset_name: Optional[str] = None,
        dev: bool = False,
        refresh_cache: bool = False
    ):
        self.dataset_name = (
            self.__class__.__name__ if dataset_name is None else dataset_name
        )
        self.root = root
        self.dev = dev
        self.triples = []
        self.samples = []
        self.task_name = "Null"
        self.entity_num = 0
        self.relation_num = 0
        self.entity2id = None
        self.relation2id = None
        self.refresh_cache = refresh_cache

        # hash filename for cache
        args_to_hash = [self.dataset_name, root] + ["dev" if dev else "prod"]
        filename = hash_str("+".join([str(arg) for arg in args_to_hash]))
        self.filepath = os.path.join(MODULE_CACHE_PATH, filename)
        self.raw_graph_process()

        self.len = len(self.triples)


    def __str__(self):
        """Prints some information of the dataset."""
        return f"Base dataset {self.dataset_name}"

    def __len__(self):
        return self.len

    def raw_graph_process(self):
        """Process the raw graph to triples (a list of triple)
        """
        raise NotImplementedError

    @staticmethod
    def info():
        """Prints the output format."""
        print(INFO_MSG)

    def stat(self):
        """Returns some statistics of the base dataset."""
        lines = list()
        lines.append("")
        lines.append(f"Statistics of base dataset (dev={self.dev}):")
        lines.append(f"\t- Dataset: {self.dataset_name}")
        lines.append(f"\t- Number of triples: {len(self.triples)}")
        lines.append(f"\t- Number of entities: {self.entity_num}")
        lines.append(f"\t- Number of relations: {self.relation_num}")
        lines.append(f"\t- Task name: {self.task_name}")
        lines.append(f"\t- Number of samples: {len(self.samples)}")
        lines.append("")
        print("\n".join(lines))
        return 

    
    def set_task(
        self,
        task_fn: Callable,
        task_name: Optional[str] = None,
        save: bool = True,
        **kwargs
    ) -> SampleKGDataset:
        """Processes the base dataset to generate the task-specific sample dataset.

        This function should be called by the user after the base dataset is
        initialized. It will iterate through all patients in the base dataset
        and call `task_fn` which should be implemented by the specific task.

        Args:
            task_fn: a function that takes a single patient and returns a
                list of samples (each sample is a dict with patient_id, visit_id,
                and other task-specific attributes as key). The samples will be
                concatenated to form the sample dataset.
            task_name: the name of the task. If None, the name of the task
                function will be used.

        Returns:
            sample_dataset: the task-specific sample (Base) dataset.

        """

        if task_name is None:
            self.task_name = task_fn.__name__

        # check if cache exists or refresh_cache is True
        if os.path.exists(self.filepath + '.pkl') and (not self.refresh_cache):
            # load from cache
            print(
                f"Loading {self.dataset_name} base dataset from {self.filepath}.pkl"
            )
            self.samples = load_pickle(self.filepath + ".pkl")
        
        else:
            print(f"Processing {self.dataset_name} base dataset...")
            self.samples = task_fn(self.triples)

            # save to cache
            print(f"Saving {self.dataset_name} base dataset to {self.filepath}")
            if save == True:
                save_pickle(self.samples, self.filepath + ".pkl")

        sample_dataset = SampleKGDataset(
            samples=self.samples,
            dataset_name=self.dataset_name,
            task_name=self.task_name,
            dev=self.dev,
            entity_num=self.entity_num,
            relation_num=self.relation_num,
            entity2id=self.entity2id,
            relation2id=self.relation2id,
            **kwargs
        )

        return sample_dataset



from typing import Dict, Optional, Tuple, Union, Callable
import os
import logging

from abc import ABC
import pandas as pd
from pandarallel import pandarallel

from pyhealth.datasets.utils import hash_str, MODULE_CACHE_PATH
from pyhealth.datasets.sample_dataset import SampleSignalDataset
from pyhealth.utils import load_pickle, save_pickle

logger = logging.getLogger(__name__)

INFO_MSG = """
dataset.patients:
    - key: patient id
    - value: recoding file paths
"""


class BaseSignalDataset(ABC):
    """Abstract base Signal dataset class.

    This abstract class defines a uniform interface for all EEG datasets
    (e.g., SleepEDF, SHHS).

    Each specific dataset will be a subclass of this abstract class, which can then
    be converted to samples dataset for different tasks by calling `self.set_task()`.

    Args:
        dataset_name: name of the dataset.
        root: root directory of the raw data (should contain many csv files).
        dev: whether to enable dev mode (only use a small subset of the data).
            Default is False.
        refresh_cache: whether to refresh the cache; if true, the dataset will
            be processed from scratch and the cache will be updated. Default is False.
    """

    def __init__(
        self,
        root: str,
        dataset_name: Optional[str] = None,
        dev: bool = False,
        refresh_cache: bool = False,
        **kwargs: Optional[Dict],
    ):

        # base attributes
        self.dataset_name = (
            self.__class__.__name__ if dataset_name is None else dataset_name
        )
        self.root = root
        self.dev = dev
        self.refresh_cache = refresh_cache

        # hash filename for cache
        args_to_hash = [self.dataset_name, root] + ["dev" if dev else "prod"]
        filename = hash_str("+".join([str(arg) for arg in args_to_hash]))
        self.filepath = os.path.join(MODULE_CACHE_PATH, filename)

        # for task-specific attributes
        self.kwargs = kwargs

        self.patients = self.process_EEG_data()

    def __str__(self):
        """Prints some information of the dataset."""
        return f"Base dataset {self.dataset_name}"

    def stat(self) -> str:
        """Returns some statistics of the base dataset."""
        lines = list()
        lines.append("")
        lines.append(f"Statistics of base dataset (dev={self.dev}):")
        lines.append(f"\t- Dataset: {self.dataset_name}")
        lines.append(f"\t- Number of patients: {len(self.patients)}")
        num_records = [len(p) for p in self.patients.values()]
        lines.append(f"\t- Number of recodings: {sum(num_records)}")
        lines.append("")
        print("\n".join(lines))
        return "\n".join(lines)

    @staticmethod
    def info():
        """Prints the output format."""
        print(INFO_MSG)

    def set_task(
        self,
        task_fn: Callable,
        task_name: Optional[str] = None,
    ) -> SampleSignalDataset:
        """Processes the base dataset to generate the task-specific sample dataset.

        This function should be called by the user after the base dataset is
        initialized. It will iterate through all patients in the base dataset
        and call `task_fn` which should be implemented by the specific task.

        Args:
            task_fn: a function that takes a single patient and returns a
                list of samples (each sample is a dict with patient_id, visit_id,
                and other task-specific attributes as key). The samples will be
                concatenated to form the sample dataset.
            task_name: the name of the task. If None, the name of the task
                function will be used.

        Returns:
            sample_dataset: the task-specific sample (Base) dataset.

        Note:
            In `task_fn`, a patient may be converted to multiple samples, e.g.,
                a patient with three visits may be converted to three samples
                ([visit 1], [visit 1, visit 2], [visit 1, visit 2, visit 3]).
                Patients can also be excluded from the task dataset by returning
                an empty list.
        """
        if task_name is None:
            task_name = task_fn.__name__

        # check if cache exists or refresh_cache is True
        if os.path.exists(self.filepath) and (not self.refresh_cache):
            """
            It obtains the signal samples (with path only) to ```self.filepath.pkl``` file
            """
            # load from cache
            logger.debug(
                f"Loaded {self.dataset_name} base dataset from {self.filepath}"
            )
            samples = load_pickle(self.filepath + ".pkl")
        else:
            """
            It stores the actual data and label to ```self.filepath/``` folder
            It also stores the signal samples (with path only) to ```self.filepath.pkl``` file
            """
            # load from raw data
            logger.debug(f"Processing {self.dataset_name} base dataset...")

            pandarallel.initialize(progress_bar=False)

            # transform dict to pandas dataframe
            if not os.path.exists(self.filepath):
                os.makedirs(self.filepath)
            patients = pd.DataFrame(self.patients.items(), columns=["pid", "records"])
            patients.records = patients.records.parallel_apply(lambda x: task_fn(x))

            samples = []
            for _, records in patients.values:
                samples.extend(records)

            # save to cache
            logger.debug(f"Saved {self.dataset_name} base dataset to {self.filepath}")
            save_pickle(samples, self.filepath + ".pkl")

        sample_dataset = SampleSignalDataset(
            samples,
            dataset_name=self.dataset_name,
            task_name=task_name,
        )
        return sample_dataset



from dataclasses import dataclass, field
from typing import Dict

from pyhealth.tasks.task_template import TaskTemplate


@dataclass(frozen=True)
class COVID19CXRClassification(TaskTemplate):
    task_name: str = "COVID19CXRClassification"
    input_schema: Dict[str, str] = field(default_factory=lambda: {"path": "image"})
    output_schema: Dict[str, str] = field(default_factory=lambda: {"label": "label"})

    def __call__(self, patient):
        return [patient]


if __name__ == "__main__":
    task = COVID19CXRClassification()
    print(task)
    print(type(task))


from abc import ABC
from pyhealth.datasets import SampleBaseDataset

import torch
import time
import numpy as np
import torch.nn as nn
import torch.nn.functional as F


class KGEBaseModel(ABC, nn.Module):
    """ Abstract class for Knowledge Graph Embedding models.

    Args:
        e_num: the number of entities in the dataset.
        r_num: the number of relations in the dataset.
        e_dim: the hidden embedding size for entity, 500 by default.
        r_dim: the hidden embedding size for relation, 500 by default.
        ns: negative sampling technique to use: can be "uniform", "normal" or "adv" (self-adversarial).
        gamma: fixed margin (only need when ns="adv").
        use_subsampling_weight: whether to use subsampling weight (like in word2vec) or not, False by default.
        use_regularization: whether to apply regularization or not, False by default.
        mode: evaluation metric type, one of "binary", "multiclass", or "multilabel", "multiclass" by default

    """

    @property
    def device(self):
        """Gets the device of the model."""
        return self._dummy_param.device


    def __init__(
        self, 
        dataset: SampleBaseDataset,
        e_dim: int = 500,
        r_dim: int = 500,
        ns: str = "uniform",
        gamma: float = None,
        use_subsampling_weight: bool = False,
        use_regularization: str = None,
        mode: str = "multiclass"
    ):
        super(KGEBaseModel, self).__init__()
        self.e_num = dataset.entity_num
        self.r_num = dataset.relation_num
        self.e_dim = e_dim
        self.r_dim = r_dim
        self.ns = ns
        self.eps = 2.0
        self.use_subsampling_weight = use_subsampling_weight
        self.use_regularization = use_regularization
        self.mode = mode
        
        if gamma != None:
            self.margin = nn.Parameter(torch.Tensor([gamma]), requires_grad=False)

        # used to query the device of the model
        self._dummy_param = nn.Parameter(torch.empty(0))


        self.E_emb = nn.Parameter(torch.zeros(self.e_num, self.e_dim))
        self.R_emb = nn.Parameter(torch.zeros(self.r_num, self.r_dim))

        if ns == "adv":
            self.e_emb_range = nn.Parameter(
                torch.Tensor([(self.margin.item() + self.eps) / e_dim]), requires_grad=False
            )

            self.r_emb_range = nn.Parameter(
                torch.Tensor([(self.margin.item() + self.eps) / r_dim]), requires_grad=False
            )

            nn.init.uniform_(
                tensor=self.E_emb, a=-self.e_emb_range.item(), b=self.e_emb_range.item()
            )

            nn.init.uniform_(
                tensor=self.R_emb, a=-self.r_emb_range.item(), b=self.r_emb_range.item()
            )

        elif ns == "normal":
            nn.init.xavier_normal_(tensor=self.E_emb)
            nn.init.xavier_normal_(tensor=self.R_emb)
        
        ## ns == "uniform"
        else:
            nn.init.xavier_uniform_(tensor=self.E_emb)
            nn.init.xavier_uniform_(tensor=self.R_emb)

    
    def data_process(self, sample_batch, mode):
        """ Data process function which converts the batch data batch into a batch of head, relation, tail

        Args:
            mode: 
                (1) 'pos': for possitive samples  
                (2) 'head': for negative samples with head prediction
                (3) 'tail' for negative samples with tail prediction
            sample_batch: 
                (1) If mode is 'pos', the sample_batch will be in shape of (batch_size, 3) where the 1-dim are 
                    triples of positive_sample in the format [head, relation, tail]
                (2) If mode is 'head' or 'tail', the sample_batch will be in shape of (batch size, 2) where the 1-dim are
                    tuples of (positive_sample, negative_sample), where positive_sample is a triple [head, relation, tail]
                    and negative_sample is a 1-d array (length: e_num) with negative (head or tail) entities indecies filled
                    and positive entities masked.

        Returns:
            head:   torch.Size([batch_size, 1, e_dim]) for tail prediction 
                    or torch.Size([batch_size, negative_sample_size(e_num), e_dim]) for head prediction
            relation: torch.Size([batch_size, 1, r_dim])
            tail:   torch.Size([batch_size, 1, e_dim]) for head prediction 
                    or torch.Size([batch_size, negative_sample_size(e_num), e_dim]) for tail prediction

        
        """
        
        if mode == "head" or mode == "tail":
            positive, negative = sample_batch
            batch_size, negative_sample_size = negative.size(0), negative.size(1)
        else:
            positive = sample_batch

        head_index = negative.view(-1) if mode == 'head' else positive[:, 0]
        tail_index = negative.view(-1) if mode == 'tail' else positive[:, 2]

        head_ = torch.index_select(self.E_emb, dim=0, index=head_index)
        head = head_.view(batch_size, negative_sample_size, -1) if mode == 'head' else head_.unsqueeze(1)

        relation = self.R_emb[positive[:, 1]].unsqueeze(1)

        tail_ = torch.index_select(self.E_emb, dim=0, index=tail_index)
        tail = tail_.view(batch_size, negative_sample_size, -1) if mode == 'tail' else tail_.unsqueeze(1)

        return head, relation, tail

    
    def train_neg_sample_gen(self, gt_head, gt_tail, negative_sampling):
        """
        (only run in train batch) 
        This function creates negative triples for training (sampling size: negative_sampling)
             with ground truth masked.
        """
        negative_sample_head = []
        negative_sample_tail = []
        for i in range(len(gt_head)):
            # head, relation, tail = triples[i]
            
            ## negative samples for head prediction
            negative_sample_list_head = []
            negative_sample_size_head = 0

            while negative_sample_size_head < negative_sampling:
                negative_sample = np.random.randint(self.e_num, size=negative_sampling*2)
                mask = np.in1d(
                    negative_sample,
                    gt_head[i],
                    assume_unique=True,
                    invert=True
                )
                negative_sample = negative_sample[mask]
                negative_sample_list_head.append(negative_sample)
                negative_sample_size_head += negative_sample.size
            
            ## negative samples for tail prediction
            negative_sample_list_tail = []
            negative_sample_size_tail = 0

            while negative_sample_size_tail < negative_sampling:
                negative_sample = np.random.randint(self.e_num, size=negative_sampling*2)
                mask = np.in1d(
                    negative_sample,
                    gt_tail[i],
                    assume_unique=True,
                    invert=True
                )
                negative_sample = negative_sample[mask]
                negative_sample_list_tail.append(negative_sample)
                negative_sample_size_tail += negative_sample.size

            neg_head = torch.LongTensor(np.concatenate(negative_sample_list_head)[:negative_sampling])
            neg_tail = torch.LongTensor(np.concatenate(negative_sample_list_tail)[:negative_sampling])
            negative_sample_head.append(neg_head)
            negative_sample_tail.append(neg_tail)
        
        negative_sample_head = torch.stack([d for d in negative_sample_head], dim=0)
        negative_sample_tail = torch.stack([d for d in negative_sample_tail], dim=0)

        return negative_sample_head, negative_sample_tail


    def test_neg_sample_filter_bias_gen(self, triples, gt_head, gt_tail):
        """
        (only run in val/test batch) 
        This function creates negative triples for validation/testing with ground truth masked.
        """

        negative_sample_head = []
        negative_sample_tail = []
        filter_bias_head = []
        filter_bias_tail = []

        for i in range(len(triples)):
            head, _, tail = triples[i]
            gt_h_ = gt_head[i]
            gt_h = gt_h_[:]
            gt_h.remove(head)
            gt_t_ = gt_tail[i]
            gt_t = gt_t_[:]
            gt_t.remove(tail)

            neg_head = np.arange(0, self.e_num)
            neg_head[gt_h] = head
            neg_head = torch.LongTensor(neg_head)

            neg_tail = np.arange(0, self.e_num)
            neg_tail[gt_t] = tail
            neg_tail = torch.LongTensor(neg_tail)

            fb_head = np.zeros(self.e_num)
            fb_head[gt_h] = -1
            fb_head = torch.LongTensor(fb_head)

            fb_tail = np.zeros(self.e_num)
            fb_tail[gt_t] = -1
            fb_tail = torch.LongTensor(fb_tail)

            negative_sample_head.append(neg_head)
            negative_sample_tail.append(neg_tail)
            filter_bias_head.append(fb_head)
            filter_bias_tail.append(fb_tail)

        negative_sample_head = torch.stack([d for d in negative_sample_head], dim=0)
        negative_sample_tail = torch.stack([d for d in negative_sample_tail], dim=0)
        filter_bias_head = torch.stack([d for d in filter_bias_head], dim=0)
        filter_bias_tail = torch.stack([d for d in filter_bias_tail], dim=0)

        return negative_sample_head, negative_sample_tail, filter_bias_head, filter_bias_tail


    def calc(self, head, relation, tail, mode='pos'):
        """ score calculation
        Args:
            head:       head entity h
            relation:   relation    r
            tail:       tail entity t
            mode: 
                (1) 'pos': for possitive samples  
                (2) 'head': for negative samples with head prediction
                (3) 'tail' for negative samples with tail prediction
        
        Return:
            score of positive/negative samples, in shape of braodcasted result of calulation with head, tail and relation.
            Example: 
                For a head prediction, suppose we have:
                    head:   torch.Size([16, 9737, 600])
                    rel:    torch.Size([16, 1, 300])
                    tail:   torch.Size([16, 1, 600])

                The unnormalized score will be in shape:
                    score:  torch.Size(16, 9737, 300)
                
                and the normalized score (return value) will be:
                    score:  torch.Size(16, 9737)
                
        """
        raise NotImplementedError


    def forward(self, **data):

        positive_sample = torch.stack([torch.LongTensor(d) for d in data['triple']], dim=0).to(self.device)

        if data['train'][0]:
            negative_sample_head, negative_sample_tail = self.train_neg_sample_gen(
                gt_head=data['ground_truth_head'],
                gt_tail=data['ground_truth_tail'],
                negative_sampling=data['hyperparameters'][0]['negative_sampling']
            )

            negative_sample_head, negative_sample_tail = negative_sample_head.to(self.device), negative_sample_tail.to(self.device)
            
            head, relation, tail = self.data_process((positive_sample, negative_sample_head), mode="head")
            neg_score_head = self.calc(head=head, relation=relation, tail=tail, mode="head")
            head, relation, tail = self.data_process((positive_sample, negative_sample_tail), mode="tail")
            neg_score_tail = self.calc(head=head, relation=relation, tail=tail, mode="tail")

            neg_score = neg_score_head + neg_score_tail

            if self.ns == 'adv':
                neg_score = (F.softmax(neg_score * 1.0, dim=1).detach() * F.logsigmoid(-neg_score)).sum(dim=1)

            else:
                neg_score = F.logsigmoid(-neg_score).mean(dim=1)

            head, relation, tail = self.data_process((positive_sample), mode="pos")
            pos_score = F.logsigmoid(self.calc(head=head, relation=relation, tail=tail)).squeeze(dim=1)

            subsampling_weight = torch.cat([d for d in data['subsampling_weight']], dim=0).to(self.device)
            pos_sample_loss = \
                - (subsampling_weight * pos_score).sum()/subsampling_weight.sum() if self.use_subsampling_weight else (- pos_score.mean())
            neg_sample_loss = \
                - (subsampling_weight * neg_score).sum()/subsampling_weight.sum() if self.use_subsampling_weight else (- neg_score.mean())

            loss = (pos_sample_loss + neg_sample_loss) / 2

            if self.use_regularization == 'l3':
                loss = loss + self.l3_regularization()
            elif self.use_regularization != None:
                loss = loss + self.regularization()

            return {"loss": loss}

        else: # valid/test
            inputs = self.test_neg_sample_filter_bias_gen(
                    triples=data['triple'],
                    gt_head=data['ground_truth_head'],
                    gt_tail=data['ground_truth_tail']
                )

            # inputs, mode = (data['positive_sample'], data['negative_sample'], data['filter_bias']), data['mode']
            inputs = [x.to(self.device) for x in inputs]
            negative_sample_head, negative_sample_tail, filter_bias_head, filter_bias_tail = inputs
            head, relation, tail = self.data_process((positive_sample, negative_sample_head), mode="head")
            score_head = self.calc(head=head, relation=relation, tail=tail, mode="head")
            head, relation, tail = self.data_process((positive_sample, negative_sample_tail), mode="tail")
            score_tail = self.calc(head=head, relation=relation, tail=tail, mode="tail")
            score_head += filter_bias_head
            score_tail += filter_bias_tail

            score = score_head + score_tail
            loss = (-F.logsigmoid(-score).mean(dim=1)).mean()
            
            
            y_true_head = positive_sample[:, 0]
            y_true_tail = positive_sample[:, 2]

            y_true = torch.cat((y_true_head, y_true_tail))
            y_prob = torch.cat((score_head, score_tail))

            return {
                "loss": loss,
                "y_true": y_true,
                "y_prob": y_prob
                }
    
    def inference(self, head=None, relation=None, tail=None, top_k=1):
        # Check if two or more arguments are None
        if sum(arg is None for arg in (head, relation, tail)) >= 2:
            print("At least 2 place holders need to be filled. ")
            return
        
        mode = "head" if head is None else ("tail" if tail is None else ("relation" if relation is None else "clf"))

        if mode == "head":
            tail_index = torch.tensor(tail)
            relation_index = torch.tensor(relation)
            relation = torch.index_select(self.R_emb, dim=0, index=relation_index).unsqueeze(1)
            tail = torch.index_select(self.E_emb, dim=0, index=tail_index).unsqueeze(1)
            head_all_idx = torch.tensor(np.arange(0, self.e_num))
            head_all = torch.index_select(self.E_emb, dim=0, index=head_all_idx).unsqueeze(1)
            score_head = self.calc(head=head_all, relation=relation, tail=tail, mode="head")
            result_eid = torch.topk(score_head.flatten(), top_k).indices
            return result_eid.tolist()

        if mode == "tail":
            head_index = torch.tensor(head)
            relation_index = torch.tensor(relation)
            head = torch.index_select(self.E_emb, dim=0, index=head_index).unsqueeze(1)
            relation = torch.index_select(self.R_emb, dim=0, index=relation_index).unsqueeze(1)
            tail_all_idx = torch.tensor(np.arange(0, self.e_num))
            tail_all = torch.index_select(self.E_emb, dim=0, index=tail_all_idx).unsqueeze(1)
            score_tail = self.calc(head=head, relation=relation, tail=tail_all, mode="tail")
            result_eid = torch.topk(score_tail.flatten(), top_k).indices
            return result_eid.tolist()
        
        if mode == "relation":
            print("Not implemented yet.")

        if mode == "clf":
            head_index = torch.tensor(head)
            relation_index = torch.tensor(relation)
            tail_index = torch.tensor(tail)
            head = torch.index_select(self.E_emb, dim=0, index=head_index).unsqueeze(1)
            relation = torch.index_select(self.R_emb, dim=0, index=relation_index).unsqueeze(1)
            tail = torch.index_select(self.E_emb, dim=0, index=tail_index).unsqueeze(1)
            score = self.calc(head=head, relation=relation, tail=tail, mode="pos")
            return score.tolist()

    
    def from_pretrained(self, path):
        state_dict = torch.load(path, map_location=self.device)
        self.update_embedding_size(state_dict)
        self.load_state_dict(state_dict)
        return

    def update_embedding_size(self, state_dict):
        e_emb_key = 'E_emb'
        r_emb_key = 'R_emb'
        
        if e_emb_key in state_dict and r_emb_key in state_dict:
            _, new_e_dim = state_dict[e_emb_key].shape
            _, new_r_dim = state_dict[r_emb_key].shape
            
            if new_e_dim != self.e_dim or new_r_dim != self.r_dim:
                self.e_dim = new_e_dim
                self.r_dim = new_r_dim
                
                self.E_emb = nn.Parameter(torch.zeros(self.e_num, self.e_dim))
                self.R_emb = nn.Parameter(torch.zeros(self.r_num, self.r_dim))
        return


            












import datetime
import logging

import gspread
from oauth2client.service_account import ServiceAccountCredentials
from google.cloud import storage
import io
import pandas as pd

from bokeh.models import HoverTool, CDSView
from bokeh.layouts import row, column, gridplot
from bokeh.models import (
    ColumnDataSource,
    CDSView,
    CheckboxGroup,
    CustomJS,
    BooleanFilter,
    Select,
    Div
)
from bokeh.plotting import figure, show, curdoc
import numpy as np
import random

from pyhealth.datasets import MIMIC3Dataset, eICUDataset, MIMIC4Dataset, OMOPDataset
from pyhealth.tasks import *
from pyhealth.metrics import *

from torch.utils.data import DataLoader


# connect to the Google Spreadsheet
def get_leaderboard_sheet(credential_file, doc_name, worksheet_id):
    scopes = [
        "https://www.googleapis.com/auth/spreadsheets",
        "https://www.googleapis.com/auth/drive",
    ]

    credentials = ServiceAccountCredentials.from_json_keyfile_name(
        credential_file, scopes
    )  # access the json key you downloaded earlier
    file = gspread.authorize(credentials)  # authenticate the JSON key with gspread
    sheet = file.open(doc_name)  # open sheet
    sheet = sheet.get_worksheet_by_id(
        worksheet_id
    )  # replace sheet_name with the name that corresponds to yours
    return sheet


# function to save the leaderboard data locally
def save_leaderboard_log(out_path, dataset_task_name, data, models):
    filename = (
        out_path + "/" + str(datetime.date.today()) + "-" + dataset_task_name + ".log"
    )
    logging.basicConfig(filename=filename, level=logging.INFO)
    logging.info(models)
    logging.info(data)


def get_dataset(dataset_name):
    dataset = None

    if dataset_name == "mimic3":
        mimic3dataset = MIMIC3Dataset(
            root="/srv/local/data/physionet.org/files/mimiciii/1.4",
            tables=["DIAGNOSES_ICD", "PROCEDURES_ICD", "PRESCRIPTIONS"],
            dev=False,
            code_mapping={"NDC": "ATC"},
            refresh_cache=False,
        )
        dataset = mimic3dataset

    elif dataset_name == "eicu":
        eicudataset = eICUDataset(
            root="/srv/local/data/physionet.org/files/eicu-crd/2.0",
            tables=["diagnosis", "medication", "physicalExam"],
            dev=False,
            refresh_cache=False,
        )
        dataset = eicudataset

    elif dataset_name == "mimic4":
        mimic4dataset = MIMIC4Dataset(
            root="/srv/local/data/physionet.org/files/mimiciv/2.0/hosp",
            tables=["diagnoses_icd", "procedures_icd", "prescriptions"],
            dev=False,
            code_mapping={"NDC": "ATC"},
            refresh_cache=False,
        )
        dataset = mimic4dataset

    elif dataset_name == "omop":
        omopdataset = OMOPDataset(
            root="/srv/local/data/zw12/pyhealth/raw_data/synpuf1k_omop_cdm_5.2.2",
            tables=[
                "condition_occurrence",
                "procedure_occurrence",
                "drug_exposure",
                "measurement",
            ],
            dev=False,
            refresh_cache=False,
        )
        dataset = omopdataset

    return dataset


def get_data_location_on_sheet(eval_data_task):
    end_line = 2 + len(eval_data_task) - 1
    location = f"A2:E{end_line}"

    return location


def get_filtered_models(all_models, remove_models):
    for rm_model in remove_models:
        if rm_model in all_models:
            all_models.remove(rm_model)

    return all_models


def get_tasks_fn_for_datasets():
    tasks_mimic3 = [
        drug_recommendation_mimic3_fn,
        length_of_stay_prediction_mimic3_fn,
        mortality_prediction_mimic3_fn,
        readmission_prediction_mimic3_fn,
    ]

    tasks_mimic4 = [
        drug_recommendation_mimic4_fn,
        length_of_stay_prediction_mimic4_fn,
        mortality_prediction_mimic4_fn,
        readmission_prediction_mimic4_fn,
    ]

    tasks_eicu = [
        drug_recommendation_eicu_fn,
        length_of_stay_prediction_eicu_fn,
        mortality_prediction_eicu_fn,
        readmission_prediction_eicu_fn,
    ]

    tasks_omop = [
        drug_recommendation_omop_fn,
        length_of_stay_prediction_omop_fn,
        mortality_prediction_omop_fn,
        readmission_prediction_omop_fn,
    ]
    return tasks_mimic3, tasks_mimic4, tasks_eicu, tasks_omop


def get_metrics_result(mode, y_gt, y_prob):
    metrics = []
    metrics_fn = None

    if mode == "multilabel":
        metrics_fn = multilabel_metrics_fn
        metrics = ["jaccard_macro", "accuracy", "f1_macro", "pr_auc_macro"]

    elif mode == "binary":
        metrics_fn = binary_metrics_fn
        metrics = ["jaccard", "accuracy", "f1", "pr_auc"]

    elif mode == "multiclass":
        metrics_fn = multiclass_metrics_fn
        metrics = ["jaccard_macro", "accuracy", "f1_macro"]

    results = metrics_fn(y_gt, y_prob, metrics=metrics, threshold=0.5)

    jaccard = results["jaccard"] if ("jaccard" in metrics) else results["jaccard_macro"]
    accuracy = results["accuracy"]
    f1 = results["f1"] if ("f1" in metrics) else results["f1_macro"]
    prauc = (
        results["pr_auc"]
        if ("pr_auc" in metrics)
        else results["pr_auc_macro"]
        if ("pr_auc_macro" in metrics)
        else "-"
    )

    # print metric name and score
    print("jaccard: ", jaccard)
    print("accuracy: ", accuracy)
    print("f1: ", f1)
    print("prauc: ", prauc)
    print("\n\n")

    return jaccard, accuracy, f1, prauc


def only_upper(s):
    return "".join(c for c in s if c.isupper())


def split_dataset_and_get_dataloaders(dataset, split_fn, ratio, collate_fn_dict):
    train_dataset, val_dataset, test_dataset = split_fn(dataset, ratio)

    train_loader = DataLoader(
        train_dataset, batch_size=64, shuffle=True, collate_fn=collate_fn_dict
    )
    val_loader = DataLoader(
        val_dataset, batch_size=64, shuffle=False, collate_fn=collate_fn_dict
    )
    test_loader = DataLoader(
        test_dataset, batch_size=64, shuffle=False, collate_fn=collate_fn_dict
    )

    return train_loader, val_loader, test_loader


def train_process(trainer, model, train_loader, val_loader, val_metric):
    try:

        trainer.train(
            train_dataloader=train_loader,
            epochs=50,
            val_dataloader=val_loader,
            monitor=val_metric,
        )

        return True

    except:
        return False


def read_dataframes_by_time_from_gcp(credentials):
    import os

    os.environ["GOOGLE_APPLICATION_CREDENTIALS"] = credentials

    dfs = {}

    storage_client = storage.Client()

    bucket_name = "pyhealth"
    prefix = "leaderboard_data/data/"
    blobs = storage_client.list_blobs(bucket_name, prefix=prefix)

    cnt = 0
    for blob in blobs:
        if cnt > 0:
            data = blob.download_as_bytes()
            df = pd.read_csv(io.BytesIO(data))
            name_ = blob.name.split("-")
            time = name_[1] + name_[2] + name_[3][:-4]
            dfs[time] = df
        cnt += 1

    return dfs


def read_dataframes_by_time_from_gcp_with_no_credentials():
    from datetime import datetime, date, timedelta

    first_data_time = datetime.strptime("2022-10-28", "%Y-%m-%d")
    now = datetime.now()
    weeks = int(abs(now - first_data_time).days / 7) + 1
    data = []
    for week in range(weeks):
        data_time = (first_data_time + timedelta(weeks=week)).strftime("%Y-%m-%d")
        file_name = f"https://storage.googleapis.com/pyhealth/leaderboard_data/data/DATA-{data_time}.csv"
        data.append(file_name)

    dfs = {}

    for d in data:
        try:
            df = pd.read_csv(d)
            name_ = d.split("-")
            time = name_[1] + name_[2] + name_[3][:-4]
            dfs[time] = df
        except:
            break

    return dfs


def get_spec_df_with_time(dfs, dataset, task):
    from datetime import datetime

    for key in dfs.keys():
        dfs[key]["date"] = key

    df = pd.concat(dfs.values())
    df["date"] = [datetime.strptime(date, "%Y%m%d") for date in df["date"].iloc()]

    if dataset == "" and task == "":
        return df

    df = df[df["Dataset-Task-Model"].str.contains(dataset)]
    df = df[df["Dataset-Task-Model"].str.contains(task)]

    return df


def make_bokeh_plot(source, metric, f):
    p = figure(
        height=190,
        width=500,
        tools=["pan, box_zoom, reset, save, crosshair"],
        toolbar_location="above",
        y_range=[-0.19, 1.2],
        # x_axis_label="date",
        y_axis_label=metric,
        x_axis_type="datetime",
    )

    if metric == "Jaccard":
        p.triangle(
            source=source,
            x="date",
            y="Jaccard",
            color="color",
            size=8,
            alpha=0.5,
            view=CDSView(source=source, filters=[f]),
        )
    if metric == "Accuracy":
        p.circle(
            source=source,
            x="date",
            y="Accuracy",
            color="color",
            size=8,
            alpha=0.5,
            view=CDSView(source=source, filters=[f]),
        )
    if metric == "F1":
        p.square(
            source=source,
            x="date",
            y="F1",
            color="color",
            size=8,
            alpha=0.5,
            view=CDSView(source=source, filters=[f]),
        )
    if metric == "PRAUC":
        p.circle(
            source=source,
            x="date",
            y="PRAUC",
            color="color",
            size=8,
            alpha=0.5,
            view=CDSView(source=source, filters=[f]),
        )
        p.xaxis.axis_label = "date"
        p.height = 220

    hover = HoverTool(
        tooltips=[
            ("model", "@{Dataset-Task-Model}"),
            ("date", "$x{%F}"),
            ("value", f"@{metric}"),
        ],
        formatters={"$x": "datetime", "Dataset-Task-Model": "printf"},
    )

    p.add_tools(hover)

    p.legend.location = "bottom_right"
    # p.xaxis.major_label_orientation = 3.14 / 2
    p.yaxis.axis_label_text_font_size = "11pt"
    p.xaxis.axis_label_text_font_size = "11pt"
    p.axis.axis_label_text_font_style = "bold"
    p.margin = (5, 5, 0, 10)

    return p


def dropdown(items, init, label):
    dp = Select(options=items, value=init, title=label)
    return dp


def generate_bokeh_figure(df):
    curdoc().theme = "caliber"

    df["letter"] = df["Dataset-Task-Model"]

    number_of_colors = 256
    colors = np.array(
        [
            "#" + "".join([random.choice("0123456789ABCDEF") for j in range(6)])
            for i in range(number_of_colors)
        ]
    )
    df["color"] = np.zeros(len(df))

    df = df.rename(columns={"Macro-F1": "F1"})

    models = sorted(set(df["Dataset-Task-Model"]))

    for model in models:
        color = random.choice(colors)
        df.loc[df["Dataset-Task-Model"].str.contains(model), "color"] = color

    dt = pd.DataFrame()
    dt["model"] = models
    for model in models:
        dataset, task = model.split("-")[0], model.split("-")[1]
        dt.loc[dt["model"].str.contains(model), "dataset"] = dataset
        dt.loc[dt["model"].str.contains(model), "task"] = task

    sc = ColumnDataSource(df)
    sc_o = ColumnDataSource(dt)
    dt_d = dt[dt["dataset"] == "mimic3"]
    sc_d = ColumnDataSource(dt_d)
    dt_t = dt[dt["task"] == "DrugRec"]
    sc_t = ColumnDataSource(dt_t)

    dp_dataset = dropdown(
        items=list(dt["dataset"].unique()), init="mimic3", label="Dataset"
    )
    dp_task = dropdown(items=list(dt["task"].unique()), init="DrugRec", label="Task")
    # dp_models = dropdown(models, "Models")

    # prefix = dp_dataset.value + '-' + dp_task.value
    # dp_models.options = [model for model in models if model.startswith(prefix)]

    init_labels = [
        *set(list(sc_t.data["model"])).intersection(list(sc_d.data["model"]))
    ]
    # active_letter = df['Dataset-Task-Model'].iloc()[0]
    f = BooleanFilter(booleans=[l == init_labels[0] for l in df["Dataset-Task-Model"]])
    cg = CheckboxGroup(labels=init_labels, active=[models.index(init_labels[0])])
    cg.js_on_change(
        "active",
        CustomJS(
            args=dict(source=sc, f=f, sc_d=sc_d, sc_t=sc_t),
            code="""\
                                const letters = cb_obj.active.map(idx => cb_obj.labels[idx]);
                                f.booleans = source.data.letter.map(l => letters.includes(l));
                                source.change.emit();
                             """,
        ),
    )

    callback_ds = CustomJS(
        args=dict(source=sc_o, sc_d=sc_d, sc_t=sc_t, other=cg),
        code="""\
                            sc_d.data['task'] = [];
                            sc_d.data['model'] = [];
                            sc_d.data['dataset'] = [];
                            for(var i = 0; i <= source.get_length(); i++){
                                if(source.data['dataset'][i] == cb_obj.value){
                                    sc_d.data['task'].push(source.data['task'][i]);
                                    sc_d.data['model'].push(source.data['model'][i]);
                                    sc_d.data['dataset'].push(source.data['dataset'][i]);
                                }
                            }
                            const intersect = sc_d.data['model'].filter(x => sc_t.data['model'].indexOf(x) !== -1);
                            other.labels = intersect;
                            other.active = new Array()
                            other.change.emit();
                            sc_d.change.emit();
                        """,
    )
    callback_tk = CustomJS(
        args=dict(source=sc_o, sc_t=sc_t, sc_d=sc_d, other=cg),
        code="""\
                            sc_t.data['model'] = [];
                            sc_t.data['task'] = [];
                            sc_t.data['dataset'] = [];
                            for(var i = 0; i <= source.get_length(); i++){
                                if(source.data['task'][i] == cb_obj.value){
                                    sc_t.data['task'].push(source.data['task'][i]);
                                    sc_t.data['model'].push(source.data['model'][i]);
                                    sc_t.data['dataset'].push(source.data['dataset'][i]);
                                }
                            }
                            const intersect = sc_t.data['model'].filter(x => sc_d.data['model'].indexOf(x) !== -1);
                            other.labels = intersect;
                            other.active = new Array()
                            other.change.emit();
                            sc_t.change.emit();
                        """,
    )
    dp_dataset.js_on_change("value", callback_ds)
    dp_task.js_on_change("value", callback_tk)
    dp_dataset.js_link("value", cg, "labels")
    dp_dataset.js_link("value", cg, "active")
    dp_task.js_link("value", cg, "labels")
    dp_task.js_link("value", cg, "active")

    div = Div(
        text="""Models (run by <a href="https://pyhealth.readthedocs.io/en/latest/api/models.html">pyhealth.models</a>)""",
        width=200,
        height=20,
    )

    cg.width = 220
    cg.height = 700
    div.margin = (20, 0, 0, 5)
    cg.margin = (5, 0, 0, 5)
    dp_task.width = 160
    dp_dataset.width = 160

    p_jac = make_bokeh_plot(sc, "Jaccard", f)
    p_acc = make_bokeh_plot(sc, "Accuracy", f)
    p_f1 = make_bokeh_plot(sc, "F1", f)
    p_prauc = make_bokeh_plot(sc, "PRAUC", f)

    # return row(column(dp_dataset, dp_task, cg), gridplot([[p_jac, p_acc], [p_f1, p_prauc]]))
    return row(
        column(dp_dataset, dp_task, div, cg), column(p_jac, p_acc, p_f1, p_prauc)
    )



import copy
import math
import random
from typing import Dict, List, Optional, Tuple

import numpy as np
import torch
import torch.nn as nn
import torch.nn.utils.rnn as rnn_utils
from sklearn.neighbors import kneighbors_graph

from pyhealth.datasets import SampleEHRDataset
from pyhealth.models import BaseModel, ConCareLayer, RNNLayer
from pyhealth.models.utils import get_last_visit


def random_init(dataset, num_centers, device):
    num_points = dataset.size(0)
    dimension = dataset.size(1)
    # print("random size", dataset.size())
    # print("numcenter", num_centers)

    indices = torch.tensor(
        np.array(random.sample(range(num_points), k=num_centers)), dtype=torch.long
    )

    centers = torch.gather(
        dataset, 0, indices.view(-1, 1).expand(-1, dimension).to(device=device)
    )
    return centers


# Compute for each data point the closest center
def compute_codes(dataset, centers):
    num_points = dataset.size(0)
    dimension = dataset.size(1)
    num_centers = centers.size(0)

    # print("size:", dataset.size(), centers.size())
    # 5e8 should vary depending on the free memory on the GPU
    # Ideally, automatically ;)
    chunk_size = int(5e8 / num_centers)
    codes = torch.zeros(num_points, dtype=torch.long)
    centers_t = torch.transpose(centers, 0, 1)
    centers_norms = torch.sum(centers**2, dim=1).view(1, -1)
    for i in range(0, num_points, chunk_size):
        begin = i
        end = min(begin + chunk_size, num_points)
        dataset_piece = dataset[begin:end, :]
        dataset_norms = torch.sum(dataset_piece**2, dim=1).view(-1, 1)
        distances = torch.mm(dataset_piece, centers_t)
        distances *= -2.0
        distances += dataset_norms
        distances += centers_norms
        _, min_ind = torch.min(distances, dim=1)
        codes[begin:end] = min_ind
    return codes


# Compute new centers as means of the data points forming the clusters
def update_centers(dataset, codes, num_centers, device):
    num_points = dataset.size(0)
    dimension = dataset.size(1)
    centers = torch.zeros(num_centers, dimension, dtype=torch.float).to(device=device)
    cnt = torch.zeros(num_centers, dtype=torch.float)
    centers.scatter_add_(
        0, codes.view(-1, 1).expand(-1, dimension).to(device=device), dataset
    )
    cnt.scatter_add_(0, codes, torch.ones(num_points, dtype=torch.float))
    # Avoiding division by zero
    # Not necessary if there are no duplicates among the data points
    cnt = torch.where(cnt > 0.5, cnt, torch.ones(num_centers, dtype=torch.float))
    centers /= cnt.view(-1, 1).to(device=device)
    return centers


def cluster(dataset, num_centers, device):
    centers = random_init(dataset, num_centers, device)
    codes = compute_codes(dataset, centers)
    num_iterations = 0
    while True:
        num_iterations += 1
        centers = update_centers(dataset, codes, num_centers, device)
        new_codes = compute_codes(dataset, centers)
        # Waiting until the clustering stops updating altogether
        # This is too strict in practice
        if torch.equal(codes, new_codes):
            break
        if num_iterations > 1000:
            break
        codes = new_codes
    return centers, codes


class GraphConvolution(nn.Module):
    def __init__(self, in_features, out_features, bias=True):
        super(GraphConvolution, self).__init__()
        self.in_features = in_features
        self.out_features = out_features
        self.weight = nn.Parameter(torch.Tensor(in_features, out_features).float())
        if bias:
            self.bias = nn.Parameter(torch.Tensor(out_features).float())
        else:
            self.register_parameter("bias", None)
        self.initialize_parameters()

    def initialize_parameters(self):
        std = 1.0 / math.sqrt(self.weight.size(1))
        self.weight.data.uniform_(-std, std)
        if self.bias is not None:
            self.bias.data.uniform_(-std, std)

    def forward(self, adj, x, device):
        y = torch.mm(x.float(), self.weight.float())
        output = torch.mm(adj.float(), y.float())
        if self.bias is not None:
            return output + self.bias.float().to(device=device)
        else:
            return output


def clones(module, N):
    "Produce N identical layers."
    return nn.ModuleList([copy.deepcopy(module) for _ in range(N)])


class GRASPLayer(nn.Module):
    """GRASPLayer layer.

    Paper: Liantao Ma et al. GRASP: generic framework for health status representation learning based on incorporating knowledge from similar patients. AAAI 2021.

    This layer is used in the GRASP model. But it can also be used as a
    standalone layer.

    Args:
        input_dim: dynamic feature size.
        static_dim: static feature size, if 0, then no static feature is used.
        hidden_dim: hidden dimension of the GRASP layer, default 128.
        cluster_num: number of clusters, default 12. The cluster_num should be no more than the number of samples.
        dropout: dropout rate, default 0.5.
        block: the backbone model used in the GRASP layer ('ConCare', 'LSTM' or 'GRU'), default 'ConCare'.

    Examples:
        >>> from pyhealth.models import GRASPLayer
        >>> input = torch.randn(3, 128, 64)  # [batch size, sequence len, feature_size]
        >>> layer = GRASPLayer(64, cluster_num=2)
        >>> c = layer(input)
        >>> c.shape
        torch.Size([3, 128])
    """

    def __init__(
        self,
        input_dim: int,
        static_dim: int = 0,
        hidden_dim: int = 128,
        cluster_num: int = 2,
        dropout: int = 0.5,
        block: str = "ConCare",
    ):
        super(GRASPLayer, self).__init__()

        self.input_dim = input_dim
        self.hidden_dim = hidden_dim
        self.cluster_num = cluster_num
        self.dropout = dropout
        self.block = block

        if self.block == "ConCare":
            self.backbone = ConCareLayer(
                input_dim, static_dim, hidden_dim, hidden_dim, dropout=0
            )
        elif self.block == "GRU":
            self.backbone = RNNLayer(input_dim, hidden_dim, rnn_type="GRU", dropout=0)
        elif self.block == "LSTM":
            self.backbone = RNNLayer(input_dim, hidden_dim, rnn_type="LSTM", dropout=0)

        self.relu = nn.ReLU()
        self.tanh = nn.Tanh()
        self.sigmoid = nn.Sigmoid()
        self.dropout = nn.Dropout(dropout)
        self.weight1 = nn.Linear(self.hidden_dim, 1)
        self.weight2 = nn.Linear(self.hidden_dim, 1)
        self.GCN = GraphConvolution(self.hidden_dim, self.hidden_dim, bias=True)
        self.GCN.initialize_parameters()
        self.GCN_2 = GraphConvolution(self.hidden_dim, self.hidden_dim, bias=True)
        self.GCN_2.initialize_parameters()
        self.A_mat = None

        self.bn = nn.BatchNorm1d(self.hidden_dim)

    def sample_gumbel(self, shape, eps=1e-20):
        U = torch.rand(shape)

        return -torch.log(-torch.log(U + eps) + eps)

    def gumbel_softmax_sample(self, logits, temperature, device):
        y = logits + self.sample_gumbel(logits.size()).to(device=device)
        return torch.softmax(y / temperature, dim=-1)

    def gumbel_softmax(self, logits, temperature, device, hard=False):
        """
        ST-gumple-softmax
        input: [*, n_class]
        return: flatten --> [*, n_class] an one-hot vector
        """
        y = self.gumbel_softmax_sample(logits, temperature, device)

        if not hard:
            return y.view(-1, self.cluster_num)

        shape = y.size()
        _, ind = y.max(dim=-1)
        y_hard = torch.zeros_like(y).view(-1, shape[-1])
        y_hard.scatter_(1, ind.view(-1, 1), 1)
        y_hard = y_hard.view(*shape)
        # Set gradients w.r.t. y_hard gradients w.r.t. y
        y_hard = (y_hard - y).detach() + y
        return y_hard

    def grasp_encoder(self, input, static=None, mask=None):

        if self.block == "ConCare":
            hidden_t, _ = self.backbone(input, mask=mask, static=static)
        else:
            _, hidden_t = self.backbone(input, mask)
        hidden_t = torch.squeeze(hidden_t, 0)

        centers, codes = cluster(hidden_t, self.cluster_num, input.device)

        if self.A_mat == None:
            A_mat = np.eye(self.cluster_num)
        else:
            A_mat = kneighbors_graph(
                np.array(centers.detach().cpu().numpy()),
                20,
                mode="connectivity",
                include_self=False,
            ).toarray()

        adj_mat = torch.tensor(A_mat).to(device=input.device)

        e = self.relu(torch.matmul(hidden_t, centers.transpose(0, 1)))  # b clu_num

        scores = self.gumbel_softmax(e, temperature=1, device=input.device, hard=True)
        digits = torch.argmax(scores, dim=-1)  #  b

        h_prime = self.relu(self.GCN(adj_mat, centers, input.device))
        h_prime = self.relu(self.GCN_2(adj_mat, h_prime, input.device))

        clu_appendix = torch.matmul(scores, h_prime)

        weight1 = torch.sigmoid(self.weight1(clu_appendix))
        weight2 = torch.sigmoid(self.weight2(hidden_t))
        weight1 = weight1 / (weight1 + weight2)
        weight2 = 1 - weight1

        final_h = weight1 * clu_appendix + weight2 * hidden_t
        out = final_h
        return out

    def forward(
        self,
        x: torch.tensor,
        static: Optional[torch.tensor] = None,
        mask: Optional[torch.tensor] = None,
    ) -> torch.tensor:
        """Forward propagation.

        Args:
            x: a tensor of shape [batch size, sequence len, input_dim].
            static: a tensor of shape [batch size, static_dim].
            mask: an optional tensor of shape [batch size, sequence len], where
                1 indicates valid and 0 indicates invalid.

        Returns:
            output: a tensor of shape [batch size, fusion_dim] representing the
                patient embedding.
        """
        # rnn will only apply dropout between layers
        out = self.grasp_encoder(x, static, mask)
        out = self.dropout(out)
        return out


class GRASP(BaseModel):
    """GRASP model.

    Paper: Liantao Ma et al. GRASP: generic framework for health status representation learning based on incorporating knowledge from similar patients. AAAI 2021.

    Note:
        We use separate GRASP layers for different feature_keys.
        Currently, we automatically support different input formats:
            - code based input (need to use the embedding table later)
            - float/int based value input
        We follow the current convention for the GRASP model:
            - case 1. [code1, code2, code3, ...]
                - we will assume the code follows the order; our model will encode
                each code into a vector and apply GRASP on the code level
            - case 2. [[code1, code2]] or [[code1, code2], [code3, code4, code5], ...]
                - we will assume the inner bracket follows the order; our model first
                use the embedding table to encode each code into a vector and then use
                average/mean pooling to get one vector for one inner bracket; then use
                GRASP one the braket level
            - case 3. [[1.5, 2.0, 0.0]] or [[1.5, 2.0, 0.0], [8, 1.2, 4.5], ...]
                - this case only makes sense when each inner bracket has the same length;
                we assume each dimension has the same meaning; we run GRASP directly
                on the inner bracket level, similar to case 1 after embedding table
            - case 4. [[[1.5, 2.0, 0.0]]] or [[[1.5, 2.0, 0.0], [8, 1.2, 4.5]], ...]
                - this case only makes sense when each inner bracket has the same length;
                we assume each dimension has the same meaning; we run GRASP directly
                on the inner bracket level, similar to case 2 after embedding table

    Args:
        dataset: the dataset to train the model. It is used to query certain
            information such as the set of all tokens.
        feature_keys:  list of keys in samples to use as features,
            e.g. ["conditions", "procedures"].
        label_key: key in samples to use as label (e.g., "drugs").
        mode: one of "binary", "multiclass", or "multilabel".
        static_keys: the key in samples to use as static features, e.g. "demographics". Default is None.
                     we only support numerical static features.
        use_embedding: list of bools indicating whether to use embedding for each feature type,
            e.g. [True, False].
        embedding_dim: the embedding dimension. Default is 128.
        hidden_dim: the hidden dimension of the GRASP layer. Default is 128.
        cluster_num: the number of clusters. Default is 10. Note that batch size should be greater than cluster_num.
        **kwargs: other parameters for the GRASP layer.


    Examples:
        >>> from pyhealth.datasets import SampleEHRDataset
        >>> samples = [
        ...         {
        ...             "patient_id": "patient-0",
        ...             "visit_id": "visit-0",
        ...             "list_codes": ["505800458", "50580045810", "50580045811"],  # NDC
        ...             "list_vectors": [[1.0, 2.55, 3.4], [4.1, 5.5, 6.0]],
        ...             "list_list_codes": [["A05B", "A05C", "A06A"], ["A11D", "A11E"]],  # ATC-4
        ...             "list_list_vectors": [
        ...                 [[1.8, 2.25, 3.41], [4.50, 5.9, 6.0]],
        ...                 [[7.7, 8.5, 9.4]],
        ...             ],
        ...             "demographic": [0.0, 2.0, 1.5],
        ...             "label": 1,
        ...         },
        ...         {
        ...             "patient_id": "patient-0",
        ...             "visit_id": "visit-1",
        ...             "list_codes": [
        ...                 "55154191800",
        ...                 "551541928",
        ...                 "55154192800",
        ...                 "705182798",
        ...                 "70518279800",
        ...             ],
        ...             "list_vectors": [[1.4, 3.2, 3.5], [4.1, 5.9, 1.7], [4.5, 5.9, 1.7]],
        ...             "list_list_codes": [["A04A", "B035", "C129"]],
        ...             "list_list_vectors": [
        ...                 [[1.0, 2.8, 3.3], [4.9, 5.0, 6.6], [7.7, 8.4, 1.3], [7.7, 8.4, 1.3]],
        ...             ],
        ...             "demographic": [0.0, 2.0, 1.5],
        ...             "label": 0,
        ...         },
        ...     ]
        >>> dataset = SampleEHRDataset(samples=samples, dataset_name="test")
        >>>
        >>> from pyhealth.models import GRASP
        >>> model = GRASP(
        ...         dataset=dataset,
        ...         feature_keys=[
        ...             "list_codes",
        ...             "list_vectors",
        ...             "list_list_codes",
        ...             "list_list_vectors",
        ...         ],
        ...         label_key="label",
        ...         static_key="demographic",
        ...         use_embedding=[True, False, True, False],
        ...         mode="binary"
        ...     )
        >>>
        >>> from pyhealth.datasets import get_dataloader
        >>> train_loader = get_dataloader(dataset, batch_size=2, shuffle=True)
        >>> data_batch = next(iter(train_loader))
        >>>
        >>> ret = model(**data_batch)
        >>> print(ret)
        {
            'loss': tensor(0.6896, grad_fn=<BinaryCrossEntropyWithLogitsBackward0>),
            'y_prob': tensor([[0.4983],
                        [0.4947]], grad_fn=<SigmoidBackward0>),
            'y_true': tensor([[1.],
                        [0.]]),
            'logit': tensor([[-0.0070],
                        [-0.0213]], grad_fn=<AddmmBackward0>)
        }
        >>>

    """

    def __init__(
        self,
        dataset: SampleEHRDataset,
        feature_keys: List[str],
        label_key: str,
        mode: str,
        use_embedding: List[bool],
        static_key: Optional[str] = None,
        embedding_dim: int = 128,
        hidden_dim: int = 128,
        **kwargs,
    ):
        super(GRASP, self).__init__(
            dataset=dataset,
            feature_keys=feature_keys,
            label_key=label_key,
            mode=mode,
        )
        self.embedding_dim = embedding_dim
        self.hidden_dim = hidden_dim
        self.use_embedding = use_embedding

        # validate kwargs for GRASP layer
        if "feature_size" in kwargs:
            raise ValueError("feature_size is determined by embedding_dim")
        if len(dataset) < 12 and "cluster_num" not in kwargs:
            raise ValueError("cluster_num is required for small dataset, default 12")
        if "cluster_num" in kwargs and kwargs["cluster_num"] > len(dataset):
            raise ValueError("cluster_num must be no larger than dataset size")

        cluster_num = kwargs.get("cluster_num", 12)

        # the key of self.feat_tokenizers only contains the code based inputs
        self.feat_tokenizers = {}
        self.static_key = static_key
        self.label_tokenizer = self.get_label_tokenizer()
        # the key of self.embeddings only contains the code based inputs
        self.embeddings = nn.ModuleDict()
        # the key of self.linear_layers only contains the float/int based inputs
        self.linear_layers = nn.ModuleDict()

        self.static_dim = 0
        if self.static_key is not None:
            self.static_dim = self.dataset.input_info[self.static_key]["len"]

        self.grasp = nn.ModuleDict()
        # add feature GRASP layers
        for idx, feature_key in enumerate(self.feature_keys):
            input_info = self.dataset.input_info[feature_key]
            # sanity check
            if input_info["type"] not in [str, float, int]:
                raise ValueError(
                    "GRASP only supports str code, float and int as input types"
                )
            elif (input_info["type"] == str) and (input_info["dim"] not in [2, 3]):
                raise ValueError(
                    "GRASP only supports 2-dim or 3-dim str code as input types"
                )
            elif (input_info["type"] == str) and (use_embedding[idx] == False):
                raise ValueError(
                    "GRASP only supports embedding for str code as input types"
                )
            elif (input_info["type"] in [float, int]) and (
                input_info["dim"] not in [2, 3]
            ):
                raise ValueError(
                    "GRASP only supports 2-dim or 3-dim float and int as input types"
                )

            # for code based input, we need Type
            # for float/int based input, we need Type, input_dim
            if use_embedding[idx]:
                self.add_feature_transform_layer(feature_key, input_info)
                self.grasp[feature_key] = GRASPLayer(
                    input_dim=embedding_dim,
                    static_dim=self.static_dim,
                    hidden_dim=self.hidden_dim,
                    **kwargs,
                )
            else:
                self.grasp[feature_key] = GRASPLayer(
                    input_dim=input_info["len"],
                    static_dim=self.static_dim,
                    hidden_dim=self.hidden_dim,
                    **kwargs,
                )

        output_size = self.get_output_size(self.label_tokenizer)
        self.fc = nn.Linear(len(self.feature_keys) * self.hidden_dim, output_size)

    def forward(self, **kwargs) -> Dict[str, torch.Tensor]:
        """Forward propagation.

        The label `kwargs[self.label_key]` is a list of labels for each patient.

        Args:
            **kwargs: keyword arguments for the model. The keys must contain
                all the feature keys and the label key.

        Returns:
            A dictionary with the following keys:
                loss: a scalar tensor representing the final loss.
                y_prob: a tensor representing the predicted probabilities.
                y_true: a tensor representing the true labels.
        """
        patient_emb = []
        for idx, feature_key in enumerate(self.feature_keys):
            input_info = self.dataset.input_info[feature_key]
            dim_, type_ = input_info["dim"], input_info["type"]

            # for case 1: [code1, code2, code3, ...]
            if (dim_ == 2) and (type_ == str):
                x = self.feat_tokenizers[feature_key].batch_encode_2d(
                    kwargs[feature_key]
                )
                # (patient, event)
                x = torch.tensor(x, dtype=torch.long, device=self.device)
                # (patient, event, embedding_dim)
                x = self.embeddings[feature_key](x)
                # (patient, event)
                mask = torch.any(x !=0, dim=2)

            # for case 2: [[code1, code2], [code3, ...], ...]
            elif (dim_ == 3) and (type_ == str):
                x = self.feat_tokenizers[feature_key].batch_encode_3d(
                    kwargs[feature_key]
                )
                # (patient, visit, event)
                x = torch.tensor(x, dtype=torch.long, device=self.device)
                # (patient, visit, event, embedding_dim)
                x = self.embeddings[feature_key](x)
                # (patient, visit, embedding_dim)
                x = torch.sum(x, dim=2)
                # (patient, visit)
                mask = torch.any(x !=0, dim=2)

            # for case 3: [[1.5, 2.0, 0.0], ...]
            elif (dim_ == 2) and (type_ in [float, int]):
                x, mask = self.padding2d(kwargs[feature_key])
                # (patient, event, values)
                x = torch.tensor(x, dtype=torch.float, device=self.device)
                # (patient, event, embedding_dim)
                if self.use_embedding[idx]:
                    x = self.linear_layers[feature_key](x)
                # (patient, event)
                mask = mask.bool().to(self.device)

            # for case 4: [[[1.5, 2.0, 0.0], [1.8, 2.4, 6.0]], ...]
            elif (dim_ == 3) and (type_ in [float, int]):
                x, mask = self.padding3d(kwargs[feature_key])
                # (patient, visit, event, values)
                x = torch.tensor(x, dtype=torch.float, device=self.device)
                # (patient, visit, embedding_dim)
                x = torch.sum(x, dim=2)

                if self.use_embedding[idx]:
                    x = self.linear_layers[feature_key](x)
                # (patient, event)
                mask = mask[:, :, 0]
                mask = mask.bool().to(self.device)
            else:
                raise NotImplementedError

            if self.static_dim > 0:
                static = torch.tensor(
                    kwargs[self.static_key], dtype=torch.float, device=self.device
                )
                x = self.grasp[feature_key](x, static=static, mask=mask)
            else:
                x = self.grasp[feature_key](x, mask=mask)
            patient_emb.append(x)

        patient_emb = torch.cat(patient_emb, dim=1)
        # (patient, label_size)
        logits = self.fc(patient_emb)
        # obtain y_true, loss, y_prob
        y_true = self.prepare_labels(kwargs[self.label_key], self.label_tokenizer)
        loss = self.get_loss_function()(logits, y_true)
        y_prob = self.prepare_y_prob(logits)
        results = {
            "loss": loss,
            "y_prob": y_prob,
            "y_true": y_true,
            "logit": logits,
        }
        if kwargs.get("embed", False):
            results["embed"] = patient_emb
        return results


if __name__ == "__main__":
    from pyhealth.datasets import SampleEHRDataset

    samples = [
        {
            "patient_id": "patient-0",
            "visit_id": "visit-0",
            # "single_vector": [1, 2, 3],
            "list_codes": ["505800458", "50580045810", "50580045811"],  # NDC
            "list_vectors": [[1.0, 2.55, 3.4], [4.1, 5.5, 6.0]],
            "list_list_codes": [["A05B", "A05C", "A06A"], ["A11D", "A11E"]],  # ATC-4
            "list_list_vectors": [
                [[1.8, 2.25, 3.41], [4.50, 5.9, 6.0]],
                [[7.7, 8.5, 9.4]],
            ],
            "label": 1,
            "demographic": [1.0, 2.0, 1.3],
        },
        {
            "patient_id": "patient-0",
            "visit_id": "visit-1",
            # "single_vector": [1, 5, 8],
            "list_codes": [
                "55154191800",
                "551541928",
                "55154192800",
                "705182798",
                "70518279800",
            ],
            "list_vectors": [[1.4, 3.2, 3.5], [4.1, 5.9, 1.7], [4.5, 5.9, 1.7]],
            "list_list_codes": [["A04A", "B035", "C129"]],
            "list_list_vectors": [
                [[1.0, 2.8, 3.3], [4.9, 5.0, 6.6], [7.7, 8.4, 1.3], [7.7, 8.4, 1.3]],
            ],
            "label": 0,
            "demographic": [1.0, 2.0, 1.3],
        },
    ]

    # dataset
    dataset = SampleEHRDataset(samples=samples, dataset_name="test")

    # data loader
    from pyhealth.datasets import get_dataloader

    train_loader = get_dataloader(dataset, batch_size=2, shuffle=True)

    # model
    model = GRASP(
        dataset=dataset,
        feature_keys=[
            "list_codes",
            "list_vectors",
            "list_list_codes",
            # "list_list_vectors",
        ],
        static_key="demographic",
        label_key="label",
        use_embedding=[True, False, True],
        mode="binary",
        cluster_num=2,
    )

    # data batch
    data_batch = next(iter(train_loader))

    # try the model
    ret = model(**data_batch)
    print(ret)

    # try loss backward
    ret["loss"].backward()



import torch
import torch.nn.functional as F
from pyhealth.models import Transformer

"""Compute Chefer's Self Attention Rules for Relevance Maps """
def apply_self_attention_rules(R_ss, cam_ss):
    return torch.matmul(cam_ss, R_ss)

""" Average Attention Scores Weighed by their Gradients """
def avg_heads(cam, grad):
    # force shapes of cam and grad to be the same order
    if len(cam.size()) < 4 and len(grad.size()) < 4: # check if no averaging needed. i.e single head
        return (grad * cam).clamp(min=0)
    cam = grad * cam # elementwise mult
    cam = cam.clamp(min=0).mean(dim=1) # average across heads
    return cam.clone()

   

class CheferRelevance():
    def __init__(self, model : Transformer):
        self.model = model
       
    """ Transformer Self Attention Token Relevance Computation
    Compute the relevance of each token in the input sequence for a given class index.
    The relevance is computed using the Chefer's Self Attention Rules.
    Paper:
        Generic Attention-model Explainability for Interpreting Bi-Modal and Encoder-Decoder Transformers
        Hila Chefer, Shir Gur, Lior Wolf
        https://arxiv.org/abs/2103.15679
        Implementation based on https://github.com/hila-chefer/Transformer-Explainability

    :param model: A trained transformer model.
    :type model: Transformer

    Examples:
        >>> from pyhealth.models import Transformer
        >>> from pyhealth.datasets import MIMIC3Dataset
        >>> from pyhealth.tasks import length_of_stay_prediction_mimic3_fn
        >>> from pyhealth.interpret.methods.chefer import CheferRelevance
        >>> mimic3_ds = MIMIC3Dataset("/srv/scratch1/data/MIMIC3").set_task(length_of_stay_prediction_mimic3_fn)
        >>> model = Transformer(dataset=mimic3_ds,feature_keys=["conditions", "procedures", "drugs"], label_key="label",mode="multiclass",)
        >>> train_data, val_data, test_data = split_by_patient(sleep_ds, [0.6, 0.2, 0.2])
        >>> train_loader = get_dataloader(train_dataset, batch_size=64, shuffle=True)
        >>> val_loader = get_dataloader(val_dataset, batch_size=64, shuffle=False)
        >>> test_loader = get_dataloader(test_dataset, batch_size=1, shuffle=False)
        >>> # ... Train the model here ...
        >>> # Get Relevance Scores for Tokens
        >>> relevance = CheferRelevance(model)
        >>> data_iterator = iter(test_loader)
        >>> data = next(data_iterator)
        >>> data['class_index'] = 0 # define class index
        >>> scores = relevance.get_relevance_matrix(**data)
        >>> print(scores)
        {'conditions': tensor([[1.2210]], device='cuda:0'), 'procedures': tensor([[1.0865]], device='cuda:0'), 'drugs': tensor([[1.]], device='cuda:0')}
    """
    def get_relevance_matrix(self, **data):
        input = data
        input['register_hook'] = True
        index = data.get('class_index')

        logits = self.model(**input)['logit']
        if index == None:
            index= torch.argmax(logits, dim=-1)

        # create one_hot matrix of n x c, one_hot vecs, for graph computation
        one_hot = F.one_hot(torch.tensor(index), logits.size()[1]).float()
        one_hot = one_hot.requires_grad_(True)
        one_hot = torch.sum(one_hot.to(logits.device) * logits)
        self.model.zero_grad()
        one_hot.backward(retain_graph=True)
        feature_keys = self.model.feature_keys 

        # get how many tokens we see per modality
        num_tokens = {}
        for key in feature_keys:
            feature_transformer = self.model.transformer[key].transformer
            for block in feature_transformer:
                num_tokens[key] = block.attention.get_attn_map().shape[-1] 

        attn  = {}
        for key in feature_keys:
            R = torch.eye(num_tokens[key]).unsqueeze(0).repeat(len(input[key]), 1, 1).to(logits.device) # initialize identity matrix, but batched
            for blk in self.model.transformer[key].transformer:
                grad = blk.attention.get_attn_grad()
                cam = blk.attention.get_attn_map()
                cam = avg_heads(cam, grad)
                R += apply_self_attention_rules(R, cam).detach()
            
            attn[key] = R[:,0] # get CLS Token

        # return Rs for each feature_key
        return attn # Assume CLS token is first row of attention score matrix 











import functools
import math
from typing import Dict, List, Optional, Tuple

import numpy as np
import torch
import torch.nn as nn

from pyhealth.datasets import BaseSignalDataset
from pyhealth.models import BaseModel


class ResBlock2D(nn.Module):
    """Convolutional Residual Block 2D

    This block stacks two convolutional layers with batch normalization,
    max pooling, dropout, and residual connection.

    Args:
        in_channels: number of input channels.
        out_channels: number of output channels.
        stride: stride of the convolutional layers.
        downsample: whether to use a downsampling residual connection.
        pooling: whether to use max pooling.

    Example:
        >>> import torch
        >>> from pyhealth.models import ResBlock2D
        >>>
        >>> model = ResBlock2D(6, 16, 1, True, True)
        >>> input_ = torch.randn((16, 6, 28, 150))  # (batch, channel, height, width)
        >>> output = model(input_)
        >>> output.shape
        torch.Size([16, 16, 14, 75])
    """

    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        stride: int = 2,
        downsample: bool = True,
        pooling: bool = True,
    ):
        super(ResBlock2D, self).__init__()
        self.conv1 = nn.Conv2d(
            in_channels, out_channels, kernel_size=3, stride=stride, padding=1
        )
        self.bn1 = nn.BatchNorm2d(out_channels)
        self.relu = nn.ELU()
        self.conv2 = nn.Conv2d(out_channels, out_channels, kernel_size=3, padding=1)
        self.bn2 = nn.BatchNorm2d(out_channels)
        self.maxpool = nn.MaxPool2d(2, stride=2)
        self.downsample = nn.Sequential(
            nn.Conv2d(
                in_channels, out_channels, kernel_size=3, stride=stride, padding=1
            ),
            nn.BatchNorm2d(out_channels),
        )
        self.downsampleOrNot = downsample
        self.pooling = pooling
        self.dropout = nn.Dropout(0.5)

    def forward(self, x):
        """Forward propagation.

        Args:
            x: input tensor of shape (batch_size, in_channels, height, width).

        Returns:
            out: output tensor of shape (batch_size, out_channels, *, *).
        """
        out = self.conv1(x)
        out = self.bn1(out)
        out = self.relu(out)
        out = self.conv2(out)
        out = self.bn2(out)
        if self.downsampleOrNot:
            residual = self.downsample(x)
            out += residual
        if self.pooling:
            out = self.maxpool(out)
        out = self.dropout(out)
        return out


class ContraWR(BaseModel):
    """The encoder model of ContraWR (a supervised model, STFT + 2D CNN layers)

    Paper: Yang, Chaoqi, Danica Xiao, M. Brandon Westover, and Jimeng Sun.
    "Self-supervised eeg representation learning for automatic sleep staging."
    arXiv preprint arXiv:2110.15278 (2021).

    Note:
        We use one encoder to handle multiple channel together.

    Args:
        dataset: the dataset to train the model. It is used to query certain
            information such as the set of all tokens.
        feature_keys:  list of keys in samples to use as features,
            e.g. ["conditions", "procedures"].
        label_key: key in samples to use as label (e.g., "drugs").
        mode: one of "binary", "multiclass", or "multilabel".
        embedding_dim: the embedding dimension. Default is 128.
        hidden_dim: the hidden dimension. Default is 128.
        **kwargs: other parameters for the Deepr layer.

    Examples:
        >>> from pyhealth.datasets import SampleSignalDataset
        >>> samples = [
        ...         {
        ...             "record_id": "SC4001-0",
        ...             "patient_id": "SC4001",
        ...             "epoch_path": "/home/chaoqiy2/.cache/pyhealth/datasets/2f06a9232e54254cbcb4b62624294d71/SC4001-0.pkl",
        ...             "label": "W",
        ...         },
        ...         {
        ...             "record_id": "SC4001-1",
        ...             "patient_id": "SC4001",
        ...             "epoch_path": "/home/chaoqiy2/.cache/pyhealth/datasets/2f06a9232e54254cbcb4b62624294d71/SC4001-1.pkl",
        ...             "label": "R",
        ...         }
        ...     ]
        >>> dataset = SampleSignalDataset(samples=samples, dataset_name="test")
        >>>
        >>> from pyhealth.models import ContraWR
        >>> model = ContraWR(
        ...         dataset=dataset,
        ...         feature_keys=["signal"], # dataloader will load the signal from "epoch_path" and put it in "signal"
        ...         label_key="label",
        ...         mode="multiclass",
        ...     )
        >>>
        >>> from pyhealth.datasets import get_dataloader
        >>> train_loader = get_dataloader(dataset, batch_size=2, shuffle=True)
        >>> data_batch = next(iter(train_loader))
        >>>
        >>> ret = model(**data_batch)
        >>> print(ret)
        {
            'loss': tensor(2.8425, device='cuda:0', grad_fn=<NllLossBackward0>),
            'y_prob': tensor([[0.9345, 0.0655],
                            [0.9482, 0.0518]], device='cuda:0', grad_fn=<SoftmaxBackward0>),
            'y_true': tensor([1, 1], device='cuda:0'),
            'logit': tensor([[ 0.1472, -2.5104],
                            [2.1584, -0.7481]], device='cuda:0', grad_fn=<AddmmBackward0>)
        }
        >>>
    """

    def __init__(
        self,
        dataset: BaseSignalDataset,
        feature_keys: List[str],
        label_key: str,
        mode: str,
        embedding_dim: int = 128,
        hidden_dim: int = 128,
        n_fft: int = 128,
        **kwargs,
    ):
        super(ContraWR, self).__init__(
            dataset=dataset,
            feature_keys=feature_keys,
            label_key=label_key,
            mode=mode,
        )
        self.embedding_dim = embedding_dim
        self.hidden_dim = hidden_dim
        self.n_fft = n_fft

        # TODO: Use more tokens for <gap> for different lengths once the input has such information
        self.label_tokenizer = self.get_label_tokenizer()

        # the ContraWR encoder
        channels, emb_size = self.cal_encoder_stat()
        self.encoder = nn.Sequential(
            *[
                ResBlock2D(channels[i], channels[i + 1], 2, True, True)
                for i in range(len(channels) - 1)
            ]
        )

        output_size = self.get_output_size(self.label_tokenizer)
        # the fully connected layer
        self.fc = nn.Linear(emb_size, output_size)

    def cal_encoder_stat(self):
        """obtain the convolution encoder initialization statistics

        Note:
            We show an example to illustrate the encoder statistics.
            input x:
                - torch.Size([5, 7, 3000])
            after stft transform
                - torch.Size([5, 7, 65, 90])
            we design the first CNN (out_channels = 8)
                - torch.Size([5, 8, 16, 22])
                - here: 8 * 16 * 22 > 256, we continute the convolution
            we design the second CNN (out_channels = 16)
                - torch.Size([5, 16, 4, 5])
                - here: 16 * 4 * 5 > 256, we continute the convolution
            we design the second CNN (out_channels = 32)
                - torch.Size([5, 32, 1, 1])
                - here: 32 * 1 * 1, we stop the convolution
            output:
                - channels = [7, 8, 16, 32]
                - emb_size = 32 * 1 * 1 = 32
        """

        print(f"\n=== Input data statistics ===")
        # obtain input signal size
        signal_info = self.dataset.input_info["signal"]
        in_channels, length = signal_info["n_channels"], signal_info["length"]
        # input signal size (batch, n_channels, length)
        print(f"n_channels: {in_channels}")
        print(f"length: {length}")

        # after stft transform (batch, n_channels, freq, time_steps)
        freq = self.n_fft // 2 + 1
        time_steps = (length - self.n_fft) // (self.n_fft // 4) + 1
        print(f"=== Spectrogram statistics ===")
        print(f"n_channels: {in_channels}")
        print(f"freq_dim: {freq}")
        print(f"time_steps: {time_steps}")

        if freq < 4 or time_steps < 4:
            raise ValueError("The input signal is too short or n_fft is too small.")

        # obtain stats at each cnn layer
        channels = [in_channels]
        cur_freq_dim = freq
        cur_time_dim = time_steps

        print(f"=== Convolution Statistics ===")
        while (cur_freq_dim >= 4 and cur_time_dim >= 4) and (
            len(channels) == 1 or cur_freq_dim * cur_time_dim * channels[-1] > 256
        ):
            channels.append(2 ** (math.floor(np.log2(channels[-1])) + 1))
            cur_freq_dim = (cur_freq_dim + 1) // 4
            cur_time_dim = (cur_time_dim + 1) // 4

            print(
                f"in_channels: {channels[-2]}, out_channels: {channels[-1]}, freq_dim: {cur_freq_dim}, time_steps: {cur_time_dim}"
            )
        print()

        emb_size = cur_freq_dim * cur_time_dim * channels[-1]
        return channels, emb_size

    def torch_stft(self, X):
        """torch short time fourier transform (STFT)

        Args:
            X: (batch, n_channels, length)

        Returns:
            signal: (batch, n_channels, freq, time_steps)
        """
        signal = []
        for s in range(X.shape[1]):
            spectral = torch.stft(
                X[:, s, :],
                n_fft=self.n_fft,
                hop_length=self.n_fft // 4,
                center=False,
                onesided=True,
                return_complex=False,
            )
            signal.append(spectral)

        signal1 = torch.stack(signal)[:, :, :, :, 0].permute(1, 0, 2, 3)
        signal2 = torch.stack(signal)[:, :, :, :, 1].permute(1, 0, 2, 3)
        signal = (signal1**2 + signal2**2) ** 0.5
        return signal

    def forward(self, **kwargs) -> Dict[str, torch.Tensor]:

        """Forward propagation."""
        # concat the info within one batch (batch, channel, length)
        x = torch.tensor(
            np.array(kwargs[self.feature_keys[0]]), device=self.device
        ).float()
        # obtain the stft spectrogram (batch, channel, freq, time step)
        x_spectrogram = self.torch_stft(x)
        # final layer embedding (batch, embedding)
        emb = self.encoder(x_spectrogram).view(x.shape[0], -1)

        # (patient, label_size)
        logits = self.fc(emb)
        # obtain y_true, loss, y_prob
        y_true = self.prepare_labels(kwargs[self.label_key], self.label_tokenizer)
        loss = self.get_loss_function()(logits, y_true)
        y_prob = self.prepare_y_prob(logits)
        results = {
            "loss": loss,
            "y_prob": y_prob,
            "y_true": y_true,
            "logit": logits,
        }
        if kwargs.get("embed", False):
            results["embed"] = emb
        return results


if __name__ == "__main__":
    """
    test the ResBlock2D
    """
    # import torch

    # input_ = torch.randn((16, 6, 3, 75))  # (batch, channel, height, width)
    # model = ResBlock2D(6, 16, 1, True, True)
    # output = model(input_)
    # print("input shape: ", input_.shape)
    # print("output.shape:", output.shape)

    # """
    # test ContraWR
    # """
    # from pyhealth.datasets import split_by_patient, get_dataloader
    # from pyhealth.trainer import Trainer
    # from pyhealth.datasets import SleepEDFDataset
    # from pyhealth.tasks import sleep_staging_sleepedf_fn

    # # step 1: load signal data
    # dataset = SleepEDFDataset(
    #     root="/srv/local/data/SLEEPEDF/sleep-edf-database-expanded-1.0.0/sleep-cassette",
    #     dev=True,
    #     refresh_cache=False,
    # )

    # # step 2: set task
    # sleep_staging_ds = dataset.set_task(sleep_staging_sleepedf_fn)
    # sleep_staging_ds.stat()
    # print(sleep_staging_ds.input_info)

    # # split dataset
    # train_dataset, val_dataset, test_dataset = split_by_patient(
    #     sleep_staging_ds, [0.6, 0.2, 0.2]
    # )
    # train_dataloader = get_dataloader(train_dataset, batch_size=5, shuffle=True)
    # val_dataloader = get_dataloader(val_dataset, batch_size=5, shuffle=False)
    # test_dataloader = get_dataloader(test_dataset, batch_size=5, shuffle=False)
    # print(
    #     "loader size: train/val/test",
    #     len(train_dataset),
    #     len(val_dataset),
    #     len(test_dataset),
    # )

    # batch = next(iter(train_dataloader))

    # # step 3: define model
    # model = ContraWR(
    #     sleep_staging_ds,
    #     feature_keys=["signal"],
    #     label_key="label",
    #     mode="multiclass",
    #     n_fft=128,
    # )

    # result = model(**batch)
    # print(result)

    """
    test ContraWR 2
    """
    from pyhealth.datasets import SampleSignalDataset, get_dataloader

    samples = [
        {
            "record_id": "SC4001-0",
            "patient_id": "SC4001",
            "epoch_path": "/home/chaoqiy2/.cache/pyhealth/datasets/2f06a9232e54254cbcb4b62624294d71/SC4001-0.pkl",
            "label": "W",
        },
        {
            "record_id": "SC4001-0",
            "patient_id": "SC4001",
            "epoch_path": "/home/chaoqiy2/.cache/pyhealth/datasets/2f06a9232e54254cbcb4b62624294d71/SC4001-1.pkl",
            "label": "R",
        },
    ]

    # dataset
    dataset = SampleSignalDataset(samples=samples, dataset_name="test")

    # data loader
    from pyhealth.datasets import get_dataloader

    train_loader = get_dataloader(dataset, batch_size=2, shuffle=True)

    # model
    model = ContraWR(
        dataset=dataset,
        feature_keys=["signal"],
        label_key="label",
        mode="multiclass",
    ).to("cuda:0")

    # data batch
    data_batch = next(iter(train_loader))

    # try the model
    ret = model(**data_batch)
    print(ret)

    # try loss backward
    ret["loss"].backward()



from typing import Dict, List, Optional

import numpy as np
import sklearn.metrics as sklearn_metrics


def regression_metrics_fn(
    x: np.ndarray,
    x_rec: np.ndarray,
    metrics: Optional[List[str]] = None,
) -> Dict[str, float]:
    """Computes metrics for regression.

    User can specify which metrics to compute by passing a list of metric names.
    The accepted metric names are:
        - kl_divergence: KL divergence
        - mse: mean squared error
        - mae: mean absolute error
    If no metrics are specified, kd_div, mse, mae are computed by default.

    This function calls sklearn.metrics functions to compute the metrics. For
    more information on the metrics, please refer to the documentation of the
    corresponding sklearn.metrics functions.

    Args:
        x: True target data sample.
        x_rec: reconstructed data sample.
        metrics: List of metrics to compute. Default is ["kl_divergence", "mse", "mae"].

    Returns:
        Dictionary of metrics whose keys are the metric names and values are
            the metric values.

    Examples:
        >>> from pyhealth.metrics import binary_metrics_fn
        >>> y_true = np.array([0, 0, 1, 1])
        >>> y_prob = np.array([0.1, 0.4, 0.35, 0.8])
        >>> binary_metrics_fn(y_true, y_prob, metrics=["accuracy"])
        {'accuracy': 0.75}
    """
    if metrics is None:
        metrics = ["kl_divergence", "mse", "mae"]

    x = x.flatten()
    x_rec = x_rec.flatten()
    
    if x.shape != x_rec.shape:
        raise ValueError("x and x_rec must have the same shape.")

    output = {}
    for metric in metrics:
        if metric == "kl_divergence":
            x[x < 1e-6] = 1e-6
            x_rec[x_rec < 1e-6] = 1e-6
            x = x / np.sum(x)
            x_rec = x_rec / np.sum(x_rec)
            kl_divergence = np.sum(x_rec * np.log(x_rec / x))
            output["kl_divergence"] = kl_divergence
        elif metric == "mse":
            mse = sklearn_metrics.mean_squared_error(x, x_rec)
            output["mse"] = mse
        elif metric == "mae":
            mae = sklearn_metrics.mean_absolute_error(x, x_rec)
            output["mae"] = mae
        else:
            raise ValueError(f"Unknown metric for regression task: {metric}")
    return output


if __name__ == "__main__":
    x = np.random.random(size=10000)
    x_rec = np.random.random(size=10000)
    print(regression_metrics_fn(x, x_rec))



from typing import Union

import torch

from pyhealth.calib.utils import LogLoss

_MAX_KERNEL_VALUE = 1.


class RBFKernelMean(torch.nn.Module):
    def __init__(self, h=1.):
        super().__init__()
        self.h = h

    def forward(self, x, x1=None):
        #if x1 is None: return 1
        _, dim = x.shape
        if x1 is None:
            x1 = x
        d = torch.pow(torch.cdist(x,x1,2), 2) / (-dim * self.h)
        return torch.exp(d)
    def set_bandwidth(self, h: Union[int, float]):
        """Set the bandwidth"""
        self.h = h
    def get_bandwidth(self):
        return self.h

def KDE_classification(X: torch.Tensor, Y: torch.Tensor, kern:RBFKernelMean=None, X_pred: torch.Tensor=None,
                       weights: Union[torch.Tensor, float, int]=1., min_eps:float=1e-10, drop_max:bool=False):
    """KDE classifier.
    We will be using X and Y to estimate the density for each of the K classes.
    kern is the kernel. X_pred is the data to predict.

    Args:
        X (torch.Tensor): Data of shape (N, d)
        Y (torch.Tensor): One-hot label of shape (N, K)
        kern (RBFKernelMean, optional): kernel.
            Defaults to None (a RBFKernelMean of bandwidth=1).
        X_pred (torch.Tensor, optional): Data to predict.
            Defaults to None, which means we will perform leave-one-out prediction.
        weights (Union[torch.Tensor, float, int], optional): Weights on each data in X.
            Defaults to 1.
        drop_max (bool, optional): Whether to ignore x where K(x_0, x) = 1.
            This typically means x is just x_0.
            Defaults to False.
            If you know there are overlap between X and X_pred, you could
            set this to True for convenience.
            Note that if X_pred=None (LOO) this is automatically set to True.

    Returns:
        torch.Tensor: Probability predictions for X_pred.
    """
    if kern is None:
        kern = RBFKernelMean(h=1.)
    #Y is a one-hot representation
    if X_pred is None:
        Kijs = kern(X, X)
        drop_max = True
    else:
        if len(X_pred.shape) == 1:
            X_pred = X_pred.unsqueeze(0)
        Kijs = kern(X_pred, X)
    if drop_max:
        Kijs = torch.where(
                Kijs < _MAX_KERNEL_VALUE-min_eps,
                Kijs,
                torch.zeros((), device=Kijs.device, dtype=Kijs.dtype)
                )
    Kijs = Kijs * weights #Kijs[:, j] *= weights[j]
    Kijs = Kijs / torch.sum(Kijs, 1, keepdim=True).clip(min_eps) #K[i,j] = K(x[i], self.X[j])
    pred = torch.matmul(Kijs, Y)
    return pred

def batched_KDE_classification(X: torch.Tensor, Y: torch.Tensor, kern=None, X_pred: torch.Tensor=None,
                       weights: Union[torch.Tensor, float, int]=1., min_eps=1e-10):
    pred = []
    batch_size=32
    drop_max = False
    if X_pred is None:
        drop_max = True
        X_pred = X
    with torch.no_grad():
        for st in range(0, len(X_pred), batch_size):
            ed = min(len(X_pred), st+batch_size)
            pred.append(KDE_classification(
                X, Y, kern, X_pred[st:ed], weights, min_eps=min_eps, drop_max=drop_max))
        return torch.concat(pred)

class KDECrossEntropyLoss(torch.nn.Module):
    reduction: str
    def __init__(self, ignore_index: int = -100,
                 reduction: str = 'mean', h: float = 1.0, nclass=None) -> None:
        super(KDECrossEntropyLoss, self).__init__()
        self.ignore_index = ignore_index
        self.reduction = reduction
        self.kern = RBFKernelMean(h=h)
        self.log_loss = LogLoss(reduction=reduction)
        self.nclass = nclass

    def forward(self, input: dict, target: torch.Tensor, eval_only=False) -> torch.Tensor:
        weights = input['weights'] if 'weights' in input else 1.
        nclass = self.nclass or max(input['supp_target'].max(), target.max()).item() + 1
        supp_Y = torch.nn.functional.one_hot(input['supp_target'], nclass).float()
        if eval_only:
            pred = batched_KDE_classification(
                input['supp_embed'], supp_Y, self.kern, weights=weights, X_pred=input['pred_embed'])
        else:
            pred = KDE_classification(
                input['supp_embed'], supp_Y, self.kern, weights=weights, X_pred=input['pred_embed'])
        ret = {}
        ret['loss'] = self.log_loss(pred, target)
        ret['extra_output'] = {"prediction": pred}
        return ret



import logging
import os
from urllib.parse import urljoin
from urllib.request import urlretrieve

import pandas as pd

from pyhealth import BASE_CACHE_PATH
from pyhealth.utils import create_directory, load_pickle, load_json

BASE_URL = "https://storage.googleapis.com/pyhealth/resource/"
MODULE_CACHE_PATH = os.path.join(BASE_CACHE_PATH, "medcode")
create_directory(MODULE_CACHE_PATH)

logger = logging.getLogger(__name__)


def download_and_read_csv(filename: str, refresh_cache: bool = False) -> pd.DataFrame:
    """Reads a csv file from the pyhealth resource folder.

    This function will read the csv file from `MODULE_CACHE_PATH` if it exists.
    Otherwise, it will download the csv file from `BASE_URL` and save it to
    `MODULE_CACHE_PATH`.

    Args:
        filename: The name of the csv file.
        refresh_cache: Whether to refresh the cache. Default is False.

    Returns:
        A pandas DataFrame.
    """
    local_filepath = os.path.join(MODULE_CACHE_PATH, filename)
    online_filepath = urljoin(BASE_URL, filename)
    if (not os.path.exists(local_filepath)) or refresh_cache:
        logger.debug(f"downloading {online_filepath} to {local_filepath}")
        urlretrieve(online_filepath, local_filepath)
    return pd.read_csv(local_filepath, dtype=str)


embedding_types = ['KG/transe', 'LM/clinicalbert', 'LM/gpt3', 'LM/biogpt', 'LM/sapbert']
features = ['conditions', 'procedures', 'drugs']
for embedding_type in embedding_types:
    for feature in features:
        MODULE_CACHE_PATH_TMP = os.path.join(BASE_CACHE_PATH, "medcode", "embeddings", embedding_type, feature)
        create_directory(MODULE_CACHE_PATH_TMP)

    MODULE_CACHE_PATH_TMP = os.path.join(BASE_CACHE_PATH, "medcode", "embeddings", embedding_type, "special_tokens")
    create_directory(MODULE_CACHE_PATH_TMP)


def download_and_read_pkl(filename: str, refresh_cache: bool = False):
    """Reads a pickle file from the pyhealth resource folder.

    This function will read the pickle file from `MODULE_CACHE_PATH` if it
    exists. Otherwise, it will download the pickle file from `BASE_URL` and
    save it to `MODULE_CACHE_PATH`.

    Args:
        filename: The name of the pickle file.
        refresh_cache: Whether to refresh the cache. Default is False.

    Returns:
        The object loaded from the pickle file.
    """
    local_filepath = os.path.join(MODULE_CACHE_PATH, filename)
    online_filepath = urljoin(BASE_URL, filename)
    if (not os.path.exists(local_filepath)) or refresh_cache:
        logger.debug(f"downloading {online_filepath} to {local_filepath}")
        urlretrieve(online_filepath, local_filepath)
    return load_pickle(local_filepath)


def download_and_read_json(filename: str, refresh_cache: bool = True):
    """Reads a json file from the pyhealth resource folder.

    This function will read the json file from `MODULE_CACHE_PATH` if it exists.
    Otherwise, it will download the json file from `BASE_URL` and save it to
    `MODULE_CACHE_PATH`.

    Args:
        filename: The name of the json file.
        refresh_cache: Whether to refresh the cache. Default is False.

    Returns:
        The object loaded from the json file.
    """
    local_filepath = os.path.join(MODULE_CACHE_PATH, filename)
    online_filepath = urljoin(BASE_URL, filename)
    if (not os.path.exists(local_filepath)) or refresh_cache:
        logger.debug(f"downloading {online_filepath} to {local_filepath}")
        urlretrieve(online_filepath, local_filepath)
    return load_json(local_filepath)



from pyhealth.medcode.inner_map import InnerMap


# TODO: add convert


class ICD10CM(InnerMap):
    """10-th International Classification of Diseases, Clinical Modification."""

    def __init__(self, **kwargs):
        super(ICD10CM, self).__init__(vocabulary="ICD10CM", **kwargs)

    @staticmethod
    def standardize(code: str):
        """Standardizes ICD10CM code."""
        if "." in code:
            return code
        if len(code) <= 3:
            return code
        return code[:3] + "." + code[3:]


if __name__ == "__main__":
    code_sys = ICD10CM(refresh_cache=True)
    code_sys.stat()
    print("A00.0" in code_sys)
    print(code_sys.lookup("D50.0"))
    print(code_sys.get_ancestors("D50.0"))
    print(code_sys.get_descendants("D50"))



from typing import Dict, List, Optional

import numpy as np

from pyhealth.metrics.fairness_utils import disparate_impact, statistical_parity_difference

def fairness_metrics_fn(
    y_true: np.ndarray,
    y_prob: np.ndarray,
    sensitive_attributes: np.ndarray,
    favorable_outcome: int = 1,
    metrics: Optional[List[str]] = None,
    threshold: float = 0.5,
) -> Dict[str, float]:
    """Computes metrics for binary classification.

    User can specify which metrics to compute by passing a list of metric names.
    The accepted metric names are:
        - disparate_impact:
        - statistical_parity_difference:

    If no metrics are disparate_impact, and statistical_parity_difference are computed by default.

    Args:
        y_true: True target values of shape (n_samples,).
        y_prob: Predicted probabilities of shape (n_samples,).
        sensitive_attributes: Sensitive attributes of shape (n_samples,) where 1 is the protected group and 0 is the unprotected group.
        favorable_outcome: Label value which is considered favorable (i.e. "positive").
        metrics: List of metrics to compute. Default is ["disparate_impact", "statistical_parity_difference"].
        threshold: Threshold for binary classification. Default is 0.5.

    Returns:
        Dictionary of metrics whose keys are the metric names and values are
            the metric values.
    """
    if metrics is None:
        metrics = ["disparate_impact", "statistical_parity_difference"]

    y_pred = y_prob.copy()
    y_pred[y_pred >= threshold] = 1
    y_pred[y_pred < threshold] = 0

    output = {}
    for metric in metrics:
        if metric == "disparate_impact":
            output[metric] = disparate_impact(sensitive_attributes, y_pred, favorable_outcome)
        elif metric == "statistical_parity_difference":
            output[metric] = statistical_parity_difference(sensitive_attributes, y_pred, favorable_outcome)
        else:
            raise ValueError(f"Unknown metric for fairness: {metric}")
    return output




from pyhealth.medcode.inner_map import InnerMap


# TODO: add standardize for different formats of NDC codes


class NDC(InnerMap):
    """National Drug Code."""

    def __init__(self, **kwargs):
        super(NDC, self).__init__(vocabulary="NDC", **kwargs)


if __name__ == "__main__":
    code_sys = NDC(refresh_cache=True)
    code_sys.stat()
    print("00527051210" in code_sys)
    print(code_sys.lookup("00527051210"))
    print(code_sys.get_ancestors("00527051210"))
    print(code_sys.get_descendants("00527051210"))



import logging
import os
from tqdm import tqdm
import numpy as np
import pandas as pd
from pandarallel import pandarallel
from pyhealth.medcode.pretrained_embeddings.kg_emb.datasets import BaseKGDataset

logger = logging.getLogger(__name__)

class UMLSDataset(BaseKGDataset):
    """Base UMLS knowleddge graph dataset

    Dataset is available at https://www.nlm.nih.gov/research/umls/index.html

    Args:
        dataset_name: name of the dataset.
        root: root directory of the raw data (should contain many csv files).
        dev: whether to enable dev mode (only use a small subset of the data).
            Default is False.
        refresh_cache: whether to refresh the cache; if true, the dataset will
            be processed from scratch and the cache will be updated. Default is False.
    
    """

    def raw_graph_process(self):
        pandarallel.initialize(progress_bar=False)
        if self.dev == False:
            self.graph_path = os.path.join(self.root, "graph.txt")
        else:
            self.graph_path = os.path.join(self.root, "graph_filtered.txt")

        if os.path.exists(self.graph_path):
            logger.debug("umls knowledge graph exists and load umls")
        else:
            logger.debug("umls does not exist")

        print("Loading UMLS knowledge graph...")
        graph_df = pd.read_csv(
            self.graph_path, 
            sep='\t',
            names=['e1', 'r', 'e2']
        )

        print("Processing UMLS knowledge graph...")
        entity_list = pd.unique(graph_df[['e1', 'e2']].values.ravel('K'))
        relation_list = pd.unique(graph_df['r'].values)

        self.entity2id = {val: i for i, val in enumerate(entity_list)}
        self.relation2id = {val: i for i, val in enumerate(relation_list)}
        self.entity_num = len(self.entity2id)
        self.relation_num = len(self.relation2id)

        print("Building UMLS knowledge graph...")
        self.triples = [(self.entity2id[e1], self.relation2id[r], self.entity2id[e2]) 
                        for e1, r, e2 in tqdm(zip(graph_df['e1'], graph_df['r'], graph_df['e2']), total=graph_df.shape[0])]


        return


if __name__ == "__main__":
    dataset = UMLSDataset(
        root="https://storage.googleapis.com/pyhealth/umls/",
        dev=True,
        refresh_cache=True,
    )
    dataset.stat()
    dataset.info()